# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
import matplotlib.pyplot as plt
import pkgutil

from ..classes.planet import Planet
from ..classes.layer import Layer


def write_tvel_file(planet_or_layer, modelname="burnmanmodel", background_model=None):
    """
    Function to write input file for obspy travel time calculations.
    Note: Because density isn't defined for most 1D seismic models, densities
    are output as zeroes.  The tvel format has a column for density,
    but this column is not used by obspy for travel time calculations.

    :param planet_or_layer: Planet or layer to write out to tvel file
    :type planet_or_layer: :class:`burnman.Planet` or :class:`burnman.Layer`.
    :param filename: Filename to read to.
    :type filename: str
    :param background_model:  1D seismic model to fill in parts of planet
        (likely to be an earth model) that aren't defined by layer
        (only need when using :class:`burnman.Layer`)
    :type background_model: :class:`burnman.seismic.Seismic1DModel`
    """

    if not isinstance(planet_or_layer, (Planet, Layer)):
        raise TypeError("Input must be a Planet() or Layer() object.")

    if isinstance(planet_or_layer, Layer):
        assert background_model
        layer = planet_or_layer
        depths = background_model.internal_depth_list()

        above_layer = np.where(depths < (np.max(depths) - layer.outer_radius))[-1]
        below_layer = np.where(depths > (np.max(depths) - layer.inner_radius))[0]

        data_above = list(
            zip(
                depths[above_layer] / 1.0e3,
                background_model.v_p(depths[above_layer]) / 1.0e3,
                background_model.v_s(depths[above_layer]) / 1.0e3,
                np.zeros_like(depths[above_layer]),
            )
        )
        data_layer = list(
            zip(
                (np.max(depths) - layer.radii)[::-1] / 1.0e3,
                layer.v_p[::-1] / 1.0e3,
                layer.v_s[::-1] / 1.0e3,
                layer.density[::-1] / 1.0e3,
            )
        )
        data_below = list(
            zip(
                depths[below_layer] / 1.0e3,
                background_model.v_p(depths[below_layer]) / 1.0e3,
                background_model.v_s(depths[below_layer]) / 1.0e3,
                np.zeros_like(depths[below_layer]),
            )
        )

        data = data_above + data_layer + data_below

        header = (
            f"{layer.name}  model from BurnMan between a radius of "
            f"{str(layer.inner_radius)} and "
            f"{str(layer.outer_radius)} km \n"
            f"{background_model.__class__.__name__} "
            f"for the rest of the Earth"
        )
        with open(modelname + ".tvel", "wb") as f:
            np.savetxt(f, data, header=header, fmt="%5.2f", delimiter="\t")

    if isinstance(planet_or_layer, Planet):
        planet = planet_or_layer
        data = list(
            zip(
                (planet.radius_planet - planet.radii)[::-1] / 1.0e3,
                planet.v_p[::-1] / 1.0e3,
                planet.v_s[::-1] / 1.0e3,
                planet.density[::-1] / 1.0e3,
            )
        )

        header = (
            f"{planet.name} model from BurnMan with a radius of "
            f"{str(planet.radius_planet)} km \n"
            f"Layers of planet are "
            f'{", ".join(layer.name for layer in planet.layers)}'
        )
        with open(modelname + ".tvel", "wb") as f:
            np.savetxt(f, data, header=header, fmt="%5.2f", delimiter="\t")


def write_axisem_input(
    layers,
    modelname="burnmanmodel_foraxisem",
    axisem_ref="axisem_prem_ani_noocean.txt",
    plotting=False,
):
    """
    Writing velocities and densities to AXISEM (www.axisem.info) input file.
    The input can be a single layer, or a list of layers taken from a planet
    (planet.layers).
    Currently this function will implement explicit discontinuities between
    layers in the seismic model.
    Currently this function is only set for Earth.

    :param layers: List of layers to put in AXISEM file.
    :type layers: list of one or more :class:`burnman.Layer`
    :param modelname: Name of model, appears in name of output file.
    :type modelname: str
    :param axisem_ref: Reference file, used to copy the header
        and for the rest of the planet, in the case of a :class:`burnman.Layer`.
    :type axisem_ref: str
    :param plotting: Choose whether to show plot of the old model and replaced model.
    :type plotting: bool
    """

    if not isinstance(layers[0], Layer):
        raise TypeError("Input must be a list of Layer()")
    # Load reference input
    datastream = pkgutil.get_data("burnman", "data/input_seismic/" + axisem_ref)
    lines = [
        line.strip() for line in datastream.decode("ascii").split("\n") if line.strip()
    ]
    table = []
    for line in lines[18:]:
        numbers = np.fromstring(line, sep=" ")
        if len(numbers) > 0:
            if line[0] != "#" and line[0] != "%":
                table.append(numbers)
    table = np.array(table)
    # format is
    # radius density vpv vsv Qk Qmu vph vsh eta

    if plotting:
        plt.figure(figsize=(12, 6))
        plt.plot(table[:, 0] / 1.0e3, table[:, 2] / 1.0e3, color="g", linestyle="--")
        plt.plot(table[:, 0] / 1.0e3, table[:, 3] / 1.0e3, color="b", linestyle="--")
        plt.plot(table[:, 0] / 1.0e3, table[:, 1] / 1.0e3, color="r", linestyle="--")

    for layer in layers:
        # Cutting out range to input in Axisem reference file, and adding in
        # model values at the top and bottom of Layer.
        i_min = next(x[0] for x in enumerate(table[:, 0]) if x[1] <= layer.outer_radius)
        if table[i_min, 0] - layer.outer_radius < 0:
            table = np.insert(table, i_min, None, axis=0)
            table[i_min, 0] = layer.outer_radius

        i_max = next(x[0] for x in enumerate(table[:, 0]) if x[1] <= layer.inner_radius)

        if table[i_max, 0] - layer.inner_radius < 0:
            table = np.insert(table, i_max, None, axis=0)
            table[i_max, 0] = layer.inner_radius

        lvp, lvs, lrho = layer.evaluate(
            ["v_p", "v_s", "density"],
            radlist=table[i_min:i_max, 0],
            radius_planet=np.max(table[:, 0]),
        )

        table[i_min:i_max, 2] = lvp
        table[i_min:i_max, 6] = lvp
        table[i_min:i_max, 3] = lvs
        table[i_min:i_max, 7] = lvs
        table[i_min:i_max, 1] = lrho

    # WRITE OUT FILE
    filename = "axisem_" + modelname + ".txt"
    f = open(filename, "w")
    print("Writing " + filename + " ...")
    f.write(
        f"# Input file {modelname} for AXISEM created using BurnMan, "
        f"replacing {axisem_ref} between "
        f"{str(np.round(layer.inner_radius / 1.e3))} and "
        f"{str(np.round(layer.outer_radius / 1.e3))} km\n"
    )

    discontinuity = 0  # Number discontinuities
    f.write("NAME " + modelname + "\n")
    for line in lines[2:18]:
        f.write(line[:] + "\n")
    for i in range(len(table[:, 0])):
        if i > 0 and table[i, 0] == table[i - 1, 0]:
            discontinuity = discontinuity + 1
            f.write(
                f"#          Discontinuity   {str(discontinuity)}, "
                f"depth:    {str(np.round((6371.e3 - table[i, 0]) / 1.e3, decimals=2))}"
                " km \n"
            )

        f.write(
            f"{table[i, 0]:8.0f} {table[i, 1]:9.2f} {table[i, 2]:9.2f} "
            f"{table[i, 3]:9.2f} {table[i, 4]:9.1f} {table[i, 5]:9.1f} "
            f"{table[i, 6]:9.2f} {table[i, 7]:9.2f} {table[i, 8]:9.5f} \n"
        )

    f.close()

    if plotting:
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 2] / 1.0e3,
            color="g",
            linestyle="-",
            label="V_p",
        )
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 3] / 1.0e3,
            color="b",
            linestyle="-",
            label="V_s",
        )
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 1] / 1.0e3,
            color="r",
            linestyle="-",
            label="density",
        )

        plt.title(
            f"{filename} = {axisem_ref} replaced between "
            f"{str(layer.inner_radius / 1.e3)} and "
            f"{str(layer.outer_radius / 1.e3)} km"
        )
        plt.legend(loc="lower right")
        plt.show()


def write_mineos_input(
    layers,
    modelname="burnmanmodel_formineos",
    mineos_ref="mineos_prem_noocean.txt",
    plotting=False,
):
    """
    Writing velocities and densities to
    Mineos (https://geodynamics.org/cig/software/mineos/) input file
    Note: currently, this function only honors the discontinuities already
    in the synthetic input file, so it is best to only replace
    certain layers with burnman values

    :param layers: List of layers to put in Mineos file.
    :type layers: list of one or more :class:`burnman.Layer`
    :param modelname: Name of model, appears in name of output file.
    :type modelname: str
    :param mineos_ref: Reference file, used to copy the header
        and for the rest of the planet, in the case of a :class:`burnman.Layer`.
    :type mineos_ref: str
    :param plotting: Choose whether to show plot of the old model and replaced model.
    :type plotting: bool
    """

    if not isinstance(layers[0], Layer):
        raise TypeError("Input must be a list of Layer()")

    # Load reference input
    datastream = pkgutil.get_data("burnman", "data/input_seismic/" + mineos_ref)
    lines = [
        line.strip() for line in datastream.decode("ascii").split("\n") if line.strip()
    ]
    table = []
    for line in lines[3:]:
        numbers = np.fromstring(line, sep=" ")
        table.append(numbers)
    table = np.array(table)

    if plotting:
        plt.figure(figsize=(12, 6))
        plt.plot(table[:, 0] / 1.0e3, table[:, 2] / 1.0e3, color="g", linestyle="--")
        plt.plot(table[:, 0] / 1.0e3, table[:, 3] / 1.0e3, color="b", linestyle="--")
        plt.plot(table[:, 0] / 1.0e3, table[:, 1] / 1.0e3, color="r", linestyle="--")

    for layer in layers:
        i_min = next(x[0] for x in enumerate(table[:, 0]) if x[1] >= layer.inner_radius)
        if table[i_min, 0] - layer.inner_radius > 0:
            table[i_min, 0] = layer.inner_radius

        i_max = next(x[0] for x in enumerate(table[:, 0]) if x[1] >= layer.outer_radius)
        if table[i_max, 0] - layer.outer_radius > 0:
            table[i_max, 0] = layer.outer_radius

        lvp, lvs, lrho = layer.evaluate(
            ["v_p", "v_s", "density"],
            radlist=table[i_min:i_max, 0],
            radius_planet=np.max(table[:, 0]),
        )

        table[i_min:i_max, 2] = lvp
        table[i_min:i_max, 6] = lvp
        table[i_min:i_max, 3] = lvs
        table[i_min:i_max, 7] = lvs
        table[i_min:i_max, 1] = lrho

    # WRITE OUT FILE
    filename = "mineos_" + modelname + ".txt"
    f = open(filename, "w")
    print("Writing " + filename + " ...")
    f.write(lines[0][:-2] + " +  " + filename + "\n")
    for line in lines[1:3]:
        f.write(line[:-2] + "\n")
    for i in range(len(table[:, 0])):
        f.write(
            "%8.0f %9.2f %9.2f %9.2f %9.1f %9.1f %9.2f %9.2f %9.5f \n"
            % (
                table[i, 0],
                table[i, 1],
                table[i, 2],
                table[i, 3],
                table[i, 4],
                table[i, 5],
                table[i, 6],
                table[i, 7],
                table[i, 8],
            )
        )

    f.close()

    if plotting:
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 2] / 1.0e3,
            color="g",
            linestyle="-",
            label="V_p",
        )
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 3] / 1.0e3,
            color="b",
            linestyle="-",
            label="V_s",
        )
        plt.plot(
            table[:, 0] / 1.0e3,
            table[:, 1] / 1.0e3,
            color="r",
            linestyle="-",
            label="density",
        )

        plt.title(
            f"{filename} = {mineos_ref} replaced between "
            f"{str(layer.inner_radius / 1.e3)} and "
            f"{str(layer.outer_radius / 1.e3)} km"
        )
        plt.legend(loc="lower right")
        plt.show()



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.classes.calibrant import Calibrant
import numpy as np

"""
Walker_2002
^^^^^^^^^^^
"""


class KCl_B2(Calibrant):
    """
    The B2 KCl pressure standard reported by
    Walker (2002; https://doi.org/10.2138/am-2002-0701).
    """

    def __init__(self):
        def _pressure_Walker_KCl(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            Pth = (
                params["aK(V,T)"] + params["dK_dT_V"] * np.log(params["V_0"] / volume)
            ) * (temperature - params["T_0"])

            # Total pressure
            P = P0 + Pth

            return P

        _params_Walker_KCl = {
            "V_0": 3.22365e-05,
            "K_0": 23.7e9,
            "Kprime_0": 4.4,
            "aK(V,T)": 0.00284e9,
            "dK_dT_V": 0.00012e9,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 1.0,
        }

        Calibrant.__init__(self, _pressure_Walker_KCl, "pressure", _params_Walker_KCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2023 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a script that converts a version of the
# HeFESTo data format into the standard burnman
# format (printed to stdout)
# This script is used to generate the 2022 dataset,
# taking parameters from https://github.com/stixrude/HeFESTo_parameters_010121
# via git clone https://github.com/stixrude/HeFESTo_parameters_010121.git
# The SLB_2022 BurnMan file is then created via the commands
# python HeFESTo_to_burnman.py > ../../minerals/SLB_2022.py; black ../../minerals/SLB_2022.py
# and from . import SLB_2022 added to the minerals __init__.py


import os
import pathlib
import pandas as pd
import numpy as np
from burnman.utils.chemistry import dictionarize_formula
from itertools import groupby
import re


def all_equal(iterable):
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def rfloat(x, m=1.0):
    return round(float(x) * m, ndigits=10)


hefesto_path = "HeFESTo_parameters_010121"
mbrdir = pathlib.Path(hefesto_path)
phasedir = pathlib.Path(f"{hefesto_path}/phase")

param_dict = {}
with open("hefesto_parameter_names.txt") as file:
    lines = [line.rstrip() for line in file]
    for line in lines:
        param_dict[" ".join(line.split()[1:])] = line.split()[0]

ignored_mbrs = [
    "crst",
    "enm",
    "fapv",
    "fea",
    "fee",
    "feg",
    "flpv",
    "hem",
    "hepv",
    "hlpv",
    "hmag",
    "hppv",
    "lppv",
    "mag",
    "mgl",
    "sil",
    "wuls",
]

solution_aliases = {
    "sp": "mg_fe_aluminous_spinel",
    "pv": "bridgmanite",
    "opx": "orthopyroxene",
    "gt": "garnet",
    "ol": "olivine",
    "cf": "calcium_ferrite_structured_phase",
    "ppv": "post_perovskite",
    "c2c": "c2c_pyroxene",
    "ri": "ringwoodite",
    "plg": "plagioclase",
    "mw": "ferropericlase",
    "cpx": "clinopyroxene",
    "nal": "new_aluminous_phase",
    "il": "ilmenite",
    "wa": "wadsleyite",
}

# solutions
sol_params = {}
for f in phasedir.iterdir():
    # Do not process liquid!!
    if os.stat(f).st_size > 0 and str(f).split("/")[-1] != "liq":
        name = str(f).split("/")[-1]
        d = pd.read_csv(f, sep="\\s+")

        endmembers = list(d.columns)

        n_mbrs = len(endmembers)

        data = d.values.tolist()

        n_data = len(data)
        v0 = n_data - n_mbrs

        # Preprocess interaction matrices
        # to get rid of ignored endmembers
        use_indices = np.array(
            [i for i, mbr in enumerate(endmembers) if mbr not in ignored_mbrs]
        )
        use_endmembers = np.array(
            [mbr for mbr in endmembers if mbr not in ignored_mbrs]
        )
        n_use_mbrs = len(use_indices)

        energy_matrix = np.array(
            np.array(data)[np.ix_(use_indices, use_indices)], dtype=float
        )
        volume_matrix = np.array(
            np.array(data)[np.ix_(v0 + use_indices, use_indices)], dtype=float
        )

        if n_data == 2 * n_mbrs + 1:
            energy_interactions = []
            energy_interactions_2 = []
            volume_interactions = []
            for i in range(n_use_mbrs - 1):
                j = i + 1
                energy_interactions.append(
                    list(np.around(energy_matrix[i][j:] * 1000.0, decimals=6))
                )
                energy_interactions_2.append(
                    list(np.around(energy_matrix[i][:j] * 1000.0, decimals=6))
                )
                volume_interactions.append(
                    list(np.around(volume_matrix[i][j:] * 1.0e-6, decimals=20))
                )

            max_abs_e = np.max(
                np.abs([item for sublist in energy_interactions for item in sublist])
            )
            max_abs_v = np.max(
                np.abs([item for sublist in volume_interactions for item in sublist])
            )
            max_abs_lower_triangular_e2 = np.max(
                np.abs([item for sublist in energy_interactions_2 for item in sublist])
            )
            assert max_abs_lower_triangular_e2 < 1.0e-3

            # print(endmembers, n_mbrs)
            # print(energy_interactions)
            if max_abs_v > 1.0e-20:
                # print(volume_interactions)
                pass

            alphas = []
            mbr_sites = []
            for mbr in use_endmembers:
                with open(f"HeFESTo_parameters_010121/{mbr}") as file:
                    lines = [line.rstrip() for line in file]
                    i_laar = [i for i, line in enumerate(lines) if "Laar" in line][0]
                    alphas.append(float(lines[i_laar].split()[0]))

                    formula_raw = lines[0].split()[0]

                    # Catch for MgTs mixing
                    if mbr == "mgts":
                        formula_raw = "Mg_1Al_1Si_2O_6"

                    # Catch for plag mixing
                    if mbr == "ab":
                        formula_raw = "Na_1Al_2Si_2O_8"

                    # Catch for mt mixing
                    if mbr == "mag":
                        formula_raw = "(Vac_1Fe_1)Fe_2O_4"

                    starts = [i for i, char in enumerate(formula_raw) if char.isupper()]
                    mixstarts = [i for i, char in enumerate(formula_raw) if char == "("]
                    mixends = [i for i, char in enumerate(formula_raw) if char == ")"]

                    for i in range(len(mixstarts)):
                        ms = mixstarts[i]
                        me = mixends[i]
                        starts = [i for i in starts if i < ms or i > me]

                    starts.extend(mixstarts)
                    starts.sort()
                    sites = [
                        formula_raw[i:j] for i, j in zip(starts, starts[1:] + [None])
                    ]
                    mbr_sites.append(sites)

            n_sites = [len(sites) for sites in mbr_sites]

            assert all_equal(n_sites)

            active_sites = []
            for i in range(n_sites[0]):
                occupancies = [sites[i] for sites in mbr_sites]
                if not all_equal(occupancies):
                    active_sites.append(i)

            site_formula_strings = []
            totals = []
            for mbr_site in mbr_sites:
                totals.append([])
                site_formula_strings.append("")
                for i in active_sites:
                    # Get number of atoms on sites
                    n_atoms = [int(s) for s in re.findall(r"\d+", mbr_site[i])]

                    species_starts = [
                        i for i, char in enumerate(mbr_site[i]) if char.isupper()
                    ]
                    species_ends = [
                        i for i, char in enumerate(mbr_site[i]) if char == "_"
                    ]
                    total_atoms = np.sum(n_atoms)
                    totals[-1].append(total_atoms)

                    site_string = "["
                    if len(n_atoms) > 1:
                        species = ""
                        for j in range(len(n_atoms)):
                            species += mbr_site[i][species_starts[j] : species_ends[j]]
                            species += f"{n_atoms[j]}/{total_atoms}"
                        site_string += species
                    else:
                        site_string += mbr_site[i][species_starts[0] : species_ends[0]]

                    site_string += "]"
                    if total_atoms != 1:
                        site_string += f"{total_atoms}"

                    site_formula_strings[-1] += site_string

            for i in range(len(totals[0])):
                t = [total[i] for total in totals]
                assert all_equal(t)

            solution_is_symmetric = np.max(np.abs(np.array(alphas) - 1.0)) < 1.0e-5

            sol_params[name] = {
                "name": name,
                "n_mbrs": n_use_mbrs,
                "mbr_names": use_endmembers,
                "mbr_site_formulae": site_formula_strings,
                "alphas": alphas,
            }

            if max_abs_e < 1.0:
                sol_params[name]["solution_type"] = "IdealSolution"
            elif solution_is_symmetric:
                sol_params[name]["solution_type"] = "SymmetricRegularSolution"
            else:
                sol_params[name]["solution_type"] = "AsymmetricRegularSolution"

            if max_abs_e > 1.0:
                sol_params[name]["energy_interaction"] = energy_interactions
            if max_abs_v > 1.0e-20:
                sol_params[name]["volume_interaction"] = volume_interactions

        else:
            pass

# endmembers
mbr_params = {}
mbr_mods = {}
mbr_names = {}
for f in mbrdir.iterdir():
    if os.stat(f).st_size > 0:
        name = str(f).split("/")[-1]

        process = False
        try:
            with open(f) as file:
                lines = [line.rstrip().split() for line in file]
                if lines[-1][-1] == "prime":
                    process = True
                if lines[0][-1] == "Liquid":
                    process = False
        except IsADirectoryError:
            pass

    if process:
        form = lines[0][0].replace("_", "").replace("(", "").replace(")", "")
        formula = dictionarize_formula(form)
        full_name = "_".join(lines[0][1:])
        # formula = formula_to_string(formula)
        # print(formula)
        # print(name)
        idict = {}
        for line in lines[1:]:
            key = param_dict[" ".join(line[1:])]
            idict[key] = line[0]

        if "T_crit" not in idict:
            idict["T_crit"] = "0.00000"

        assert idict["debye"] == "1.00000"

    if idict["bel_0"] != "0.00000":
        # Fe endmember equations of state not yet implemented
        process = False

    if idict["zpp"] != "0.00000":
        process = False

    if process:
        mbr_params[name] = {
            "name": full_name,
            "formula": formula,
            "equation_of_state": "slb3",
            "F_0": rfloat(idict["F_0"], 1.0e3),
            "V_0": rfloat(idict["V_0"], 1.0e-6),
            "K_0": rfloat(idict["K_0"], 1.0e9),
            "Kprime_0": rfloat(idict["Kprime_0"]),
            "Debye_0": rfloat(idict["Theta_0"]),
            "grueneisen_0": rfloat(idict["gamma_0"]),
            "q_0": rfloat(idict["q_0"]),
            "G_0": rfloat(idict["G_0"], 1.0e9),
            "Gprime_0": rfloat(idict["Gprime_0"]),
            "eta_s_0": rfloat(idict["eta_s_0"]),
            "n": rfloat(idict["n"]),
            "Z": rfloat(idict["Z"]),
            "molar_mass": rfloat(idict["molar_mass"], 1.0e-3),
        }

        if idict["T_crit"] != "0.00000":
            mbr_mods[name] = [
                [
                    "landau_slb_2022",
                    {
                        "Tc_0": float(idict["T_crit"]),
                        "S_D": float(idict["S_crit"]),
                        "V_D": float(idict["V_crit"]),
                    },
                ]
            ]
        mbr_names[name] = (
            full_name.lower()
            .replace("-", "_")
            .replace(" ", "_")
            .replace("'", "")
            .replace('\\"', "")
        )

# WRITE FILE

print(
    "# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit\n"
    "# for the Earth and Planetary Sciences\n"
    "# Copyright (C) 2012 - 2023 by the BurnMan team,\n"
    "# released under the GNU GPL v2 or later.\n"
    "\n\n"
    '"""\n'
    "SLB_2022\n"
    "Minerals from Stixrude & Lithgow-Bertelloni 2022 and references therein\n"
    f"File autogenerated using HeFESTO_to_burnman.py and {hefesto_path}\n"
    '"""\n\n'
    "from __future__ import absolute_import\n"
    "\n\n"
    "from ..classes.mineral import Mineral\n"
    "from ..classes.solution import Solution\n"
    "from ..classes.solutionmodel import (\n"
    "    IdealSolution,\n"
    "    SymmetricRegularSolution,\n"
    "    AsymmetricRegularSolution,\n"
    ")\n\n"
)

print('"""\n' "ENDMEMBERS\n" '"""\n')
for key, mbr_prm in sorted(mbr_params.items()):
    print(
        f"\nclass {key}(Mineral):\n"
        "    def __init__(self):\n"
        f"        self.params = {mbr_prm}\n"
    )
    if key in mbr_mods:
        print(
            f"        self.property_modifiers = {mbr_mods[key]}",
        )
        print("")
    print("        Mineral.__init__(self)")
    print("")

print('"""\n' "SOLUTIONS\n" '"""\n')


for key, prm in sorted(sol_params.items()):
    docstring = '"""'
    docstring += f'{prm["solution_type"]} model for {solution_aliases[key]} ({key}).\n'
    docstring += "Endmembers (and site species distributions) are given in the order:\n"
    for i in range(prm["n_mbrs"]):
        docstring += f'- {prm["mbr_names"][i]} ({prm["mbr_site_formulae"][i]})\n'
    if key == "mw":
        docstring += "The entropy from the first site in magnetite is not counted.\n"
    docstring += '"""'

    print(
        f"\nclass {solution_aliases[key]}(Solution):\n"
        "    def __init__(self, molar_fractions=None):\n"
        f"        {docstring}\n"
        f'        self.name = "{solution_aliases[key]}"\n'
        f"        self.solution_model = {prm['solution_type']}(\n"
        "            endmembers=["
    )
    for i in range(prm["n_mbrs"]):
        print(
            f'                [{prm["mbr_names"][i]}(), "{prm["mbr_site_formulae"][i]}"],'
        )
    print("            ],")

    if prm["solution_type"] == "AsymmetricRegularSolution":
        print(f"            alphas={prm['alphas']},")

    if prm["solution_type"] != "IdealSolution":
        print(f'            energy_interaction={prm["energy_interaction"]},')
        if "volume_interaction" in prm:
            print(f'            volume_interaction={prm["volume_interaction"]},')

    print("        )\n")
    print("        Solution.__init__(self, molar_fractions=molar_fractions)\n")


print('\n"""\n' "ENDMEMBER ALIASES\n" '"""\n')

for key, value in sorted(mbr_names.items()):
    print(f"{value} = {key}")

print("")


print('\n"""\n' "SOLUTION ALIASES\n" '"""\n')

for key, value in sorted(solution_aliases.items()):
    if key != "sp":
        print(f"{key} = {value}")



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from . import equation_of_state as eos
import warnings
from math import exp


def bulk_modulus(volume, params):
    """
    compute the bulk modulus as per the
    Vinet equation of state.  Reference bulk
    modulus should be in :math:`[Pa]`.
    """

    x = volume / params["V_0"]
    eta = (3.0 / 2.0) * (params["Kprime_0"] - 1.0)

    K = (
        (params["K_0"] * pow(x, -2.0 / 3.0))
        * (1 + ((eta * pow(x, 1.0 / 3.0) + 1.0) * (1.0 - pow(x, 1.0 / 3.0))))
        * exp(eta * (1.0 - pow(x, 1.0 / 3.0)))
    )
    return K


def vinet(x, params):
    """
    equation for the  Vinet equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0']), which should be in math:`[Pa]`.
    """
    eta = (3.0 / 2.0) * (params["Kprime_0"] - 1.0)
    return (
        3.0
        * params["K_0"]
        * (pow(x, -2.0 / 3.0))
        * (1.0 - (pow(x, 1.0 / 3.0)))
        * exp(eta * (1.0 - pow(x, 1.0 / 3.0)))
        + params["P_0"]
    )


def volume(pressure, params):
    """
    Get the Vinet volume at a reference temperature for a given
    pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
    """

    func = lambda x: vinet(x / params["V_0"], params) - pressure
    V = opt.brentq(func, 0.1 * params["V_0"], 1.5 * params["V_0"])
    return V


class Vinet(eos.EquationOfState):
    """
    Base class for the isothermal Vinet equation of state.
    References for this equation of state are :cite:`vinet1986`
    and :cite:`vinet1987`. This equation of state actually
    predates Vinet by 55 years :cite:`Rydberg1932`,
    and was investigated further by :cite:`Stacey1981`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return vinet(volume / params["V_0"], params)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        Currently not included in the Vinet EOS, so omitted.
        """
        return 0.0

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = pow(volume / params["V_0"], 1.0 / 3.0)
        eta = (3.0 / 2.0) * (params["Kprime_0"] - 1.0)

        intPdV = (
            9.0
            * params["V_0"]
            * params["K_0"]
            / (eta * eta)
            * ((1.0 - eta * (1.0 - x)) * exp(eta * (1.0 - x)) - 1.0)
        )

        return -intPdV + params["E_0"]

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + volume * pressure
        )

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 0.0

        # G is not included in the Vinet EOS so we shall set them to NaN's
        if "G_0" not in params:
            params["G_0"] = float("nan")
        if "Gprime_0" not in params:
            params["Gprime_0"] = float("nan")

        # check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # now check that the values are reasonable.  I mostly just
        # made up these values from experience, and we are only
        # raising a warning.  Better way to do this? [IR]
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < -5.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.
import numpy as np
import copy
from scipy.linalg import expm, logm
from scipy.optimize import minimize
from .solution import Solution
from .anisotropicmineral import (
    AnisotropicMineral,
    convert_f_Pth_to_f_T_derivatives,
    deformation_gradient_alpha_and_compliance,
)
from .material import material_property, cached_property, Material
from ..utils.anisotropy import (
    contract_stresses,
    expand_stresses,
    voigt_notation_to_compliance_tensor,
)
from ..utils.misc import copy_documentation


class AnisotropicSolution(Solution, AnisotropicMineral):
    """
    A class implementing the anisotropic solution model described
    in :cite:`Myhill2024a`.
    This class is derived from Solution and AnisotropicMineral,
    and inherits most of the methods from those classes.

    Instantiation of an AnisotropicSolution is similar to that of a scalar
    Solution, except that each of the endmembers must be an instance of
    an AnisotropicMineral, and an additional function and parameter dictionary
    are passed to the constructor of AnisotropicSolution that describe
    excess contributions to the anisotropic state tensor (Psi_xs)
    and its derivatives with respect to volume and temperature.
    The function arguments should be ln(V), Pth,
    p (a vector of proportions) and the parameter dictionary,
    in that order.
    The output variables Psi_xs_Voigt, dPsidf_Pth_Voigt_xs and
    dPsidPth_f_Voigt_xs (all 6x6 matrices) and dPsiIdp_xs
    (a 3 x 3 x n_endmember matrix) must be returned in that
    order in a tuple.
    States of the mineral can only be queried after setting the
    pressure and temperature using set_state() and the composition using
    set_composition().

    This class is available as ``burnman.AnisotropicSolution``.
    """

    def __init__(
        self,
        name,
        solution_model,
        psi_excess_function,
        anisotropic_parameters,
        molar_fractions=None,
    ):
        # Always set orthotropic as false to ensure correct
        # derivatives are taken.
        # To do: relax this to allow faster calculations
        self.orthotropic = False

        self.T_0 = 298.15

        # Initialise as both Material and Solution object
        Material.__init__(self)
        Solution.__init__(self, name, solution_model, molar_fractions)

        # Store a scalar copy of the solution model to speed up set_state
        scalar_solution_model = copy.copy(solution_model)
        scalar_solution_model.endmembers = [
            [mbr.isotropic_mineral, formula] for mbr, formula in self.endmembers
        ]
        self._scalar_solution = Solution(name, scalar_solution_model, molar_fractions)

        self._logm_M0_mbr = np.einsum(
            "kij->ijk", np.array([logm(m[0].cell_vectors_0.T) for m in self.endmembers])
        )
        self.frame_convention = self.endmembers[0][0].frame_convention
        self.anisotropic_params = anisotropic_parameters
        self.psi_excess_function = psi_excess_function

        # Finally, set the composition
        if molar_fractions is not None:
            self.set_composition(molar_fractions)

    def set_state(self, pressure, temperature):
        # Set solution conditions
        ss = self._scalar_solution
        ss.set_state(pressure, temperature)

        try:
            self.compute_base_properties()
        except AttributeError:
            raise Exception("You must use set_composition() before set_state().")

        Material.set_state(self, pressure, temperature)

    def compute_base_properties(self):
        pressure = self._scalar_solution.pressure
        temperature = self._scalar_solution.temperature

        ss = self._scalar_solution
        V = ss.molar_volume
        KT_at_T = ss.isothermal_bulk_modulus_reuss
        f = np.log(V)
        self._f = f
        # Evaluate endmember properties at V, T
        # Here we explicitly manipulate each of the anisotropic endmembers
        pressure_guesses = [np.max([0.0e9, pressure - 2.0e9]), pressure + 2.0e9]
        mbrs = self.endmembers
        for mbr in mbrs:
            mbr[0].set_state_with_volume(V, temperature, pressure_guesses)

        # Endmember cell vectors and other functions of Psi (all unrotated)
        self._PsiI_mbr = np.einsum("kij->ijk", np.array([m[0]._PsiI for m in mbrs]))
        dPsidf_T_Voigt_mbr = np.array([m[0]._dPsidf_T_Voigt for m in mbrs])
        dPsiIdf_T_mbr = np.array([m[0]._dPsiIdf_T for m in mbrs])
        dPsiIdT_f_mbr = np.array([m[0]._dPsiIdT_f for m in mbrs])

        fr = self.molar_fractions
        PsiI_ideal = np.einsum("ijp, p->ij", self._PsiI_mbr, fr)
        dPsidf_T_Voigt_ideal = np.einsum("p,pij->ij", fr, dPsidf_T_Voigt_mbr)
        dPsiIdf_T_ideal = np.einsum("p,pij->ij", fr, dPsiIdf_T_mbr)
        dPsiIdT_f_ideal = np.einsum("p,pij->ij", fr, dPsiIdT_f_mbr)

        # Now calculate the thermal pressure terms by
        # evaluating the scalar solution at V, T_0
        ss.set_state_with_volume(V, self.T_0)
        P_at_T0 = ss.pressure
        KT_at_T0 = ss.isothermal_bulk_modulus_reuss
        self.dPthdf = KT_at_T0 - KT_at_T
        self.Pth = pressure - P_at_T0

        # And we're done with calculating endmember properties at V
        # Now we can set state of this object and the scalar one
        ss.set_state(pressure, temperature)
        Solution.set_state(self, pressure, temperature)

        # Calculate excess properties at the given f and Pth
        out = self.psi_excess_function(
            f, self.Pth, self.molar_fractions, self.anisotropic_params
        )
        Psi_xs_Voigt, dPsidf_Pth_Voigt_xs, dPsidPth_f_Voigt_xs = out[:3]
        self._dPsiIdp_xs = out[3]
        Psi_xs_full = voigt_notation_to_compliance_tensor(Psi_xs_Voigt)
        PsiI_xs = np.einsum("ijkl, kl", Psi_xs_full, np.eye(3))

        # Change of variables: (f, Pth) -> Psi(f, T)
        aK_T = ss.alpha * ss.isothermal_bulk_modulus_reuss
        out = convert_f_Pth_to_f_T_derivatives(
            dPsidf_Pth_Voigt_xs, dPsidPth_f_Voigt_xs, aK_T, self.dPthdf
        )
        dPsidf_T_Voigt_xs, dPsiIdf_T_xs, dPsiIdT_f_xs = out

        self._PsiI = PsiI_ideal + PsiI_xs

        out = deformation_gradient_alpha_and_compliance(
            ss.alpha,
            ss.isothermal_compressibility_reuss,
            self._PsiI,
            dPsidf_T_Voigt_ideal + dPsidf_T_Voigt_xs,
            dPsiIdf_T_ideal + dPsiIdf_T_xs,
            dPsiIdT_f_ideal + dPsiIdT_f_xs,
        )
        (
            self._unrotated_F,
            self._unrotated_dFdf_T,
            self._unrotated_alpha,
            self._unrotated_S_T_Voigt,
        ) = out

    def set_composition(self, molar_fractions):
        self._scalar_solution.set_composition(molar_fractions)
        self.cell_vectors_0 = expm(
            np.einsum("ijk, k ->ji", self._logm_M0_mbr, molar_fractions)
        )
        # Calculate all other required properties
        Solution.set_composition(self, molar_fractions)
        if self._scalar_solution.pressure is not None:
            self.compute_base_properties()

    @cached_property
    def ones(self):
        return np.ones(self.n_endmembers)

    @cached_property
    def eye(self):
        return np.eye(self.n_endmembers)

    @material_property
    def _dM0dp_fixed_VT(self):
        """
        Gradient in the standard state cell tensor
        with respect to molar proportions at constant
        volume and temperature under hydrostatic conditions.
        """
        lnM0_ones = np.einsum("ij, k->kij", logm(self.cell_vectors_0.T), self.ones)
        dp = 1.0e-5

        dlnM0 = np.einsum("ijk->kij", self._logm_M0_mbr) * dp / 2.0
        logmM0_0 = lnM0_ones - dlnM0
        logmM0_1 = lnM0_ones + dlnM0
        return np.einsum("kij->ijk", (expm(logmM0_1) - expm(logmM0_0)) / dp)

    @material_property
    def _unrotated_dFdp_fixed_VT(self):
        """
        Gradient in the unrotated deformation gradient tensor
        with respect to molar proportions at constant
        volume and temperature under hydrostatic conditions.
        """
        dp = 1.0e-5
        PsiI_ones = np.einsum("ij, k->kij", self._PsiI, self.ones)
        dPsiI = np.einsum("ijk->kij", self._PsiI_mbr + self._dPsiIdp_xs) * dp / 2.0
        PsiI_0 = PsiI_ones - dPsiI
        PsiI_1 = PsiI_ones + dPsiI

        return np.einsum("kij->ijk", (expm(PsiI_1) - expm(PsiI_0)) / dp)

    @material_property
    def _unrotated_dMdn_fixed_VT(self):
        """
        Gradient in unrotated cell tensor with respect to composition
        at fixed volume and temperature under hydrostatic conditions.
        """
        F = self._unrotated_F
        M0 = self.cell_vectors_0.T
        fr = self.molar_fractions
        n_dpdn = self.eye - np.einsum("l, m->lm", self.ones, fr)
        n = np.sum(self.molar_fractions)
        n23 = np.power(n, 2.0 / 3.0)

        dM0dn = (
            np.einsum("kj, l->kjl", M0, self.ones / 3.0)
            + np.einsum("kjm,lm->kjl", self._dM0dp_fixed_VT, n_dpdn)
        ) / n23
        n_dVmoldn = -self.molar_volume * self.ones

        dFdV = self._unrotated_dFdf_T / self.molar_volume
        dFdn = (
            np.einsum("ij, l->ijl", dFdV, n_dVmoldn)
            + np.einsum("ikm,lm->ikl", self._unrotated_dFdp_fixed_VT, n_dpdn)
        ) / n

        dFdnM0 = np.einsum("ikl,kj->ijl", dFdn, M0)
        FdM0dn = np.einsum("ik,kjl->ijl", F, dM0dn)
        return dFdnM0 + FdM0dn

    @material_property
    def _dMdn_fixed_VT(self):
        """
        Gradient in cell tensor with respect to composition
        at fixed volume and temperature under hydrostatic conditions.
        """
        R = self.rotation_matrix
        return np.einsum("mi, nj, ijk->mnk", R, R, self._unrotated_dMdn_fixed_VT)

    @material_property
    def depsdn_fixed_VT(self):
        """
        Gradient in strain with respect to composition
        at fixed volume and temperature under hydrostatic conditions.
        """
        invM = np.linalg.inv(self.cell_vectors.T)
        Ln = np.einsum("ijm,kj->ikm", self._dMdn_fixed_VT, invM)
        LnT = np.einsum("ikm->kim", Ln)
        return 0.5 * (Ln + LnT)


depsdeps = np.einsum("pm, qn->pqmn", np.eye(3), np.eye(3))


class RelaxedAnisotropicSolution(AnisotropicSolution):
    """
    A class implementing the relaxed anisotropic solution
    model described in :cite:`Myhill2024b`.
    This class is derived from AnisotropicSolution,
    and inherits most of the methods from that class.

    Instantiation of a RelaxedAnisotropicSolution involves passing
    an AnisotropicSolution, plus a set of vectors that represent rapid
    deformation modes. For example, a solution of MgO, FeHSO and FeLSO
    (high and low spin wuestite) can rapidly change proportion of
    high spin and low spin iron, and so a single vector should be passed:
    np.array([[0., -1., 1.]]) or some multiple thereof.

    States of the mineral can only be queried after setting the
    pressure and temperature using set_state() and the composition using
    set_composition().

    This class is available as ``burnman.RelaxedAnisotropicSolution``.
    """

    def __init__(
        self,
        anisotropic_solution,
        relaxation_vectors,
        unrelaxed_vectors,
    ):
        # Make an attribute with the unrelaxed solution
        self.unrelaxed = anisotropic_solution

        # The relaxation vectors
        self.n_relaxation_vectors = len(relaxation_vectors)
        self.n_unrelaxed_vectors = len(unrelaxed_vectors)

        self.dndq = np.array(relaxation_vectors).T
        assert len(self.dndq.shape) == 2
        assert len(self.dndq) == self.unrelaxed.n_endmembers

        self.dndx = np.array(unrelaxed_vectors).T
        assert len(self.dndx.shape) == 2
        assert len(self.dndx) == self.unrelaxed.n_endmembers
        assert (
            len(unrelaxed_vectors) + len(relaxation_vectors)
            == self.unrelaxed.n_endmembers
        )

        self.q_initial = np.zeros(len(relaxation_vectors)) + 0.001

        try:
            molar_fractions = anisotropic_solution.molar_fractions
        except AttributeError:
            molar_fractions = None

        # Give the relaxed solution the same base properties as the
        # unrelaxed solution
        AnisotropicSolution.__init__(
            self,
            name=anisotropic_solution.name,
            solution_model=anisotropic_solution.solution_model,
            psi_excess_function=anisotropic_solution.psi_excess_function,
            anisotropic_parameters=anisotropic_solution.anisotropic_params,
            molar_fractions=molar_fractions,
        )

    def set_state(self, pressure, temperature, relaxed=True):
        """
        Sets the state of the solution. Also relaxes the
        structure parameters if set_composition() has already
        been used and if the relaxed argument has been set to
        True.

        :param pressure: The pressure of the solution [Pa]
        :type pressure: float
        :param temperature: The temperature of the solution [K]
        :type temperature: float
        :param relaxed: Whether to minimize the Gibbs energy
            of the material by changing the values of the structure
            parameters. Defaults to True.
        :type relaxed: bool, optional
        """
        self.unrelaxed.set_state(pressure, temperature)

        if hasattr(self.unrelaxed, "molar_fractions") and relaxed:
            self._relax_at_PTX()
            AnisotropicSolution.set_state(self, pressure, temperature)

    def set_composition(self, molar_fractions, q_initial=None, relaxed=True):
        """
        Sets the composition of the model. Also relaxes the
        structure parameters if set_state() has already
        been used and if the relaxed argument has been set to
        True.

        :param molar_fractions: Molar fractions of the
            independent endmembers corresponding to the
            unrelaxed vectors specified during initialisation.
        :type molar_fractions: 1D numpy array
        :param q_initial: Initial values of the structure parameters.
            Defaults to None, in which case the preexisting
            initial values are used
            (first set to 0.001 during initialisation).
        :type q_initial: 1D numpy array, optional
        :param relaxed: Whether to minimize the Gibbs energy
            of the material by changing the values of the structure
            parameters. Defaults to True.
        :type relaxed: bool, optional
        """
        self.unrelaxed_vectors = molar_fractions
        if q_initial is not None:
            self.q_initial = np.array(q_initial)
        n = np.einsum("ij, j", self.dndq, self.q_initial) + np.einsum(
            "ij, j", self.dndx, molar_fractions
        )

        self.unrelaxed.set_composition(n)

        if self.unrelaxed.pressure is not None and relaxed:
            self._relax_at_PTX()
            AnisotropicSolution.set_composition(self, n)
            self.unrelaxed._scalar_solution.set_composition(self.molar_fractions)
            self.unrelaxed.set_composition(self.molar_fractions)

    def _relax_at_PTX(self):
        """
        Minimizes the Gibbs energy at constant pressure and
        temperature by changing the structural parameters.

        Run during set_state() and set_composition(), as long as both
        state and composition have already been set. This function
        should not generally be needed by the user.
        """
        n0 = copy.copy(self.unrelaxed._scalar_solution.molar_fractions)

        def G_func(dq):
            n = n0 + np.einsum("ij, j", self.dndq, dq)
            self.unrelaxed._scalar_solution.set_composition(n)
            return self.unrelaxed._scalar_solution.molar_gibbs

        sol = minimize(G_func, np.zeros(len(self.dndq[0])), method="Nelder-Mead")
        assert sol.success
        n = n0 + np.einsum("ij, j", self.dndq, sol.x)
        self.unrelaxed.set_composition(n)
        AnisotropicSolution.set_composition(self, n)

    def set_state_with_volume(
        self, volume, temperature, pressure_guesses=[0.0e9, 10.0e9]
    ):
        """
        This function acts similarly to set_state, but takes volume and
        temperature as input to find the pressure. In order to ensure
        self-consistency, this function does not use any pressure functions
        from the material classes, but instead finds the pressure using the
        brentq root-finding and Nelder-Mead minimization methods.

        Composition should have been set before this function is used.

        :param volume: The desired molar volume of the mineral [m^3].
        :type volume: float

        :param temperature: The desired temperature of the mineral [K].
        :type temperature: float

        :param pressure_guesses: A list of floats denoting the initial
            low and high guesses for bracketing of the pressure [Pa].
            These guesses should preferably bound the correct pressure,
            but do not need to do so. More importantly,
            they should not lie outside the valid region of
            the equation of state. Defaults to [0.e9, 10.e9].
        :type pressure_guesses: list
        """
        ss = self.unrelaxed._scalar_solution
        n0 = copy.copy(ss.molar_fractions)

        # Initial set_state without changing structural parameters
        # to estimate pressure
        ss.set_state_with_volume(volume, temperature, pressure_guesses)
        # Store in a mutable so that it can be updated in the loop
        pressure = [ss.pressure]

        def F_func(dq):
            n = n0 + np.einsum("ij, j", self.dndq, dq)
            ss.set_composition(n)
            ss.set_state_with_volume(
                volume, temperature, [pressure[0] - 1.0e9, pressure[0] + 1.0e9]
            )
            pressure[0] = ss.pressure
            return ss.molar_helmholtz

        sol = minimize(F_func, np.zeros(len(self.dndq[0])), method="Nelder-Mead")
        assert sol.success

        # Run one more time with root
        F_func(sol.x)
        self.unrelaxed.set_composition(ss.molar_fractions)
        AnisotropicSolution.set_composition(self, ss.molar_fractions)
        self.set_state(pressure[0], temperature)

    @material_property
    def _depsdq_fixed_VT(self):
        """
        Gradient in strain with respect to structural state
        at fixed volume and temperature under hydrostatic conditions.
        """
        return np.einsum("mnq, qu->mnu", self.depsdn_fixed_VT, self.dndq)

    @material_property
    def _depsdT_fixed_volume(self):
        """
        Gradient in strain with respect to temperature
        at fixed volume and structural state under hydrostatic conditions.
        """
        beta_T = self.unrelaxed.isothermal_compressibility_tensor
        beta_TR = self.unrelaxed.isothermal_compressibility_reuss
        alpha = self.unrelaxed.thermal_expansivity_tensor
        alpha_V = self.unrelaxed.alpha

        return (alpha / alpha_V - beta_T / beta_TR) * alpha_V

    @material_property
    def _dbarepsdeps(self):
        """
        Gradient in non-hydrostatic strain with
        respect to total strain at fixed temperature
        and structural state under hydrostatic conditions.
        """
        beta_T = self.unrelaxed.isothermal_compressibility_tensor
        beta_TR = self.unrelaxed.isothermal_compressibility_reuss
        return depsdeps - np.einsum("pq, mn->pqmn", beta_T / beta_TR, np.eye(3))

    @material_property
    def _dSdq(self):
        """
        Gradient in strain with respect to structural state
        at constant volume and temperature under hydrostatic
        conditions.

        This function converts the partial entropies
        from Gibbs natural variables (P, T)
        to Helmholtz natural variables (V, T).
        """
        dSdn = (
            self._scalar_solution.partial_entropies
            - self._scalar_solution.alpha
            * self._scalar_solution.isothermal_bulk_modulus_reuss
            * self._scalar_solution.partial_volumes
        )
        return np.einsum("i, ij->j", dSdn, self.dndq)

    @material_property
    def _dPdq(self):
        """
        Pressure gradient with respect to structural state
        calculated at constant volume and temperature
        under hydrostatic conditions.

        This function converts from Gibbs natural variables
        (partial volumes as a function of P and T)
        to Helmholtz natural variables
        (partial pressures as a function of V and T).
        """
        dPdn = (
            self._scalar_solution.isothermal_bulk_modulus_reuss
            / self._scalar_solution.V
            * self._scalar_solution.partial_volumes
        )
        return np.einsum("i, ij->j", dPdn, self.dndq)

    @material_property
    def _d2Fdqdq_fixed_VT(self):
        """
        Helmholtz structural hessian
        calculated at constant volume and temperature
        under hydrostatic conditions.

        This function converts from Gibbs natural variables
        (the Gibbs compositional hessian at constant P and T)
        to Helmholtz natural variables
        (the Helmholtz compositional hessian
        at constant V and T).
        """
        d2Fdndn = (
            self._scalar_solution.gibbs_hessian
            + np.einsum(
                "i, j->ij",
                self._scalar_solution.partial_volumes,
                self._scalar_solution.partial_volumes,
            )
            * self._scalar_solution.isothermal_bulk_modulus_reuss
            / self._scalar_solution.V
        )
        return np.einsum("ij, ik, jl->kl", d2Fdndn, self.dndq, self.dndq)

    @material_property
    def _d2Fdqdq_fixed_epsT(self):
        """
        Second structure parameter derivative of the Helmholtz energy
        at constant strain and temperature.
        """
        C_T = self.unrelaxed.full_isothermal_stiffness_tensor
        return self._d2Fdqdq_fixed_VT + self.V * np.einsum(
            "kli, klmn, mnj->ij", self._depsdq_fixed_VT, C_T, self._depsdq_fixed_VT
        )

    @material_property
    def _d2Fdqdz(self):
        """
        Second derivatives of the Helmholtz energy with
        respect to composition and z (strain or temperature).
        """
        C_T = self.unrelaxed.full_isothermal_stiffness_tensor
        d2FdqdT = -self._dSdq + self.V * np.einsum(
            "kli, klmn, mn->i", self._depsdq_fixed_VT, C_T, self._depsdT_fixed_volume
        )

        d2Fdqdeps = -self.V * (
            np.einsum("i, mn->imn", self._dPdq, np.eye(3))
            + np.einsum(
                "kli, klpq, pqmn->imn", self._depsdq_fixed_VT, C_T, self._dbarepsdeps
            )
        )
        d2Fdqdeps = np.array([contract_stresses(arr) for arr in d2Fdqdeps])

        return np.concatenate((d2Fdqdeps, d2FdqdT[:, np.newaxis]), axis=1)

    @material_property
    def _d2Fdqdq_fixed_epsT_pinv(self):
        """
        The inverse of the second derivative of the Helmholtz energy
        with respect to the structure parameters at constant
        strain and temperature. Often referred to as the
        susceptibility matrix.
        """
        return np.linalg.pinv(self._d2Fdqdq_fixed_epsT)

    @material_property
    def dqdz_relaxed(self):
        """
        The change of the structure parameters with respect to
        strain and temperature that minimizes the Helmholtz
        energy.
        """
        return -np.einsum("kl, lj->kj", self._d2Fdqdq_fixed_epsT_pinv, self._d2Fdqdz)

    @material_property
    def _d2Fdzdz_Q(self):
        """
        Block matrix of V*C_T, V*pi, -c_eps/T
        at fixed Q
        """
        CT = self.unrelaxed.isothermal_stiffness_tensor
        pi = contract_stresses(self.unrelaxed.thermal_stress_tensor)
        c_eps = self.unrelaxed.molar_isometric_heat_capacity
        V = self.molar_volume
        T = self.temperature
        return np.block([[V * CT, V * pi[:, np.newaxis]], [V * pi, -c_eps / T]])

    @material_property
    def _d2Fdzdz(self):
        """
        Block matrix of V*C_T, V*pi, -c_eps/T
        under Helmholtz-minimizing Q
        """
        return self._d2Fdzdz_Q + np.einsum(
            "ki, kj->ij", self._d2Fdqdz, self.dqdz_relaxed
        )

    # The next three functions provide the three relaxed
    # second derivative properties from which other properties
    # may be derived.
    @material_property
    @copy_documentation(AnisotropicMineral.isothermal_stiffness_tensor)
    def isothermal_stiffness_tensor(self):
        return self._d2Fdzdz[:6, :6] / self.V

    @material_property
    @copy_documentation(AnisotropicMineral.thermal_stress_tensor)
    def thermal_stress_tensor(self):
        return expand_stresses(self._d2Fdzdz[6, :6] / self.V)

    @material_property
    @copy_documentation(AnisotropicMineral.molar_isometric_heat_capacity)
    def molar_isometric_heat_capacity(self):
        return -self._d2Fdzdz[6, 6] * self.temperature

    # The following tensor properties are all
    # second derivatives that are calculated in AnisotropicMineral
    # (rather than being derived from other properties),
    # and must therefore be redefined in relaxed materials

    # The full and Voigt isothermal stiffness tensors,
    # full and contracted isentropic stiffness and compliance tensors
    # Grueneisen tensor are derived properties
    # in AnisotropicMineral
    # and so do not need to be redefined.

    @material_property
    @copy_documentation(AnisotropicMineral.isothermal_compliance_tensor)
    def isothermal_compliance_tensor(self):
        return np.linalg.pinv(self.isothermal_stiffness_tensor)

    @material_property
    @copy_documentation(AnisotropicMineral.full_isothermal_compliance_tensor)
    def full_isothermal_compliance_tensor(self):
        S_Voigt = self.isothermal_compliance_tensor
        return voigt_notation_to_compliance_tensor(S_Voigt)

    @material_property
    @copy_documentation(AnisotropicMineral.thermal_expansivity_tensor)
    def thermal_expansivity_tensor(self):
        alpha = -np.einsum(
            "ijkl, kl",
            self.full_isothermal_compliance_tensor,
            self.thermal_stress_tensor,
        )
        return alpha

    # The following scalar properties are all
    # second derivatives that are calculated in AnisotropicMineral
    # or Solution (rather than being derived from other properties),
    # and must therefore be redefined in relaxed materials

    # The volumetric molar heat capacity and grueneisen parameter are
    # derived properties in AnisotropicMineral
    # and so do not need to be redefined.

    @material_property
    @copy_documentation(AnisotropicMineral.isothermal_compressibility_reuss)
    def isothermal_compressibility_reuss(self):
        return np.trace(self.isothermal_compressibility_tensor)

    @material_property
    @copy_documentation(AnisotropicMineral.isothermal_bulk_modulus_reuss)
    def isothermal_bulk_modulus_reuss(self):
        return 1.0 / self.isothermal_compressibility_reuss

    @material_property
    @copy_documentation(AnisotropicMineral.isentropic_compressibility_reuss)
    def isentropic_compressibility_reuss(self):
        return np.trace(self.isentropic_compressibility_tensor)

    @material_property
    @copy_documentation(AnisotropicMineral.isentropic_bulk_modulus_reuss)
    def isentropic_bulk_modulus_reuss(self):
        return 1.0 / self.isentropic_compressibility_reuss

    @material_property
    @copy_documentation(AnisotropicMineral.thermal_expansivity)
    def thermal_expansivity(self):
        return np.trace(self.thermal_expansivity_tensor)

    @material_property
    @copy_documentation(AnisotropicMineral.molar_heat_capacity_p)
    def molar_heat_capacity_p(self):
        alpha = self.thermal_expansivity_tensor
        pi = self.thermal_stress_tensor
        C_p = (
            self.molar_isometric_heat_capacity
            - self.V * self.temperature * np.einsum("ij, ij", alpha, pi)
        )
        return C_p



# This file is part of BurnMan - a thermoelastic
# and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
mp50KFMASH
^^^^^^^^^^

HPx-eos solutions using endmembers from
dataset HP_2011_ds62.
The values in this document are all in S.I. units,
unlike those in the original THERMOCALC file.
This file is autogenerated using process_HPX_eos.py
"""

from numpy import array, nan
from . import HP_2011_ds62
from ..classes.mineral import Mineral
from ..classes.solution import Solution
from ..classes.solutionmodel import SymmetricRegularSolution
from ..classes.solutionmodel import AsymmetricRegularSolution
from ..classes.combinedmineral import CombinedMineral


annm = CombinedMineral([HP_2011_ds62.ann()], [1.0], [-3000.0, -0.0, 0.0], "annm")
obi = CombinedMineral(
    [HP_2011_ds62.ann(), HP_2011_ds62.phl()],
    [0.3333333333333333, 0.6666666666666666],
    [-3000.0, -0.0, 0.0],
    "obi",
)
fm = CombinedMineral(
    [HP_2011_ds62.en(), HP_2011_ds62.fs()], [0.5, 0.5], [-6600.0, -0.0, 0.0], "fm"
)
fspm = CombinedMineral([HP_2011_ds62.fspr()], [1.0], [-2000.0, -0.0, 0.0], "fspm")
spro = CombinedMineral(
    [HP_2011_ds62.fspr(), HP_2011_ds62.spr4()],
    [0.75, 0.25],
    [-3500.0, -0.0, 0.0],
    "spro",
)
mstm = CombinedMineral([HP_2011_ds62.mst()], [1.0], [0.0, -0.0, 0.0], "mstm")
ochl1 = CombinedMineral(
    [HP_2011_ds62.afchl(), HP_2011_ds62.clin(), HP_2011_ds62.daph()],
    [1.0, -1.0, 1.0],
    [3000.0, -0.0, 0.0],
    "ochl1",
)
ochl4 = CombinedMineral(
    [HP_2011_ds62.afchl(), HP_2011_ds62.clin(), HP_2011_ds62.daph()],
    [1.0, -0.2, 0.2],
    [2400.0, -0.0, 0.0],
    "ochl4",
)


class g(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "g"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.py(), "[Mgx]3"],
                [HP_2011_ds62.alm(), "[Fex]3"],
            ],
            energy_interaction=[[2500.0]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class mu(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "mu"
        self.solution_model = AsymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.mu(), "[Alma][Sit1/2Alt1/2]2"],
                [HP_2011_ds62.cel(), "[Mgma][Sit]2"],
                [HP_2011_ds62.fcel(), "[Fema][Sit]2"],
            ],
            alphas=[0.63, 0.63, 0.63],
            energy_interaction=[[0.0, 0.0], [0.0]],
            volume_interaction=[
                [2.0000000000000003e-06, 2.0000000000000003e-06],
                [0.0],
            ],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class bi(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "bi"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.phl(), "[Mgm][Mgm]2[Sit1/2Alt1/2]2"],
                [annm, "[Fem][Fem]2[Sit1/2Alt1/2]2"],
                [obi, "[Fem][Mgm]2[Sit1/2Alt1/2]2"],
                [HP_2011_ds62.east(), "[Alm][Mgm]2[Alt]2"],
            ],
            energy_interaction=[
                [12000.0, 4000.0, 10000.0],
                [8000.0, 15000.0],
                [7000.0],
            ],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class opx(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "opx"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.en(), "[Mgm][Mgm][Sit]1/2"],
                [HP_2011_ds62.fs(), "[Fem][Fem][Sit]1/2"],
                [fm, "[Mgm][Fem][Sit]1/2"],
                [HP_2011_ds62.mgts(), "[Alm][Mgm][Alt1/2Sit1/2]1/2"],
            ],
            energy_interaction=[
                [7000.0, 4000.0, 13000.0],
                [4000.0, 13000.0],
                [17000.0],
            ],
            volume_interaction=[[0.0, 0.0, -1.5e-06], [0.0, -1.5e-06], [-1.5e-06]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class sa(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "sa"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.spr4(), "[Mgm][Mgm]3[Sit]"],
                [HP_2011_ds62.spr5(), "[Alm][Mgm]3[Alt]"],
                [fspm, "[Fem][Fem]3[Sit]"],
                [spro, "[Mgm][Fem]3[Sit]"],
            ],
            energy_interaction=[
                [10000.0, 16000.0, 12000.0],
                [19000.0, 22000.0],
                [4000.0],
            ],
            volume_interaction=[
                [-2.0000000000000002e-07, 0.0, 0.0],
                [-2.0000000000000002e-07, -2.0000000000000002e-07],
                [0.0],
            ],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class cd(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "cd"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.crd(), "[Mgx]2[Vh]"],
                [HP_2011_ds62.fcrd(), "[Fex]2[Vh]"],
                [HP_2011_ds62.hcrd(), "[Mgx]2[Hoh]"],
            ],
            energy_interaction=[[8000.0, 0.0], [9000.0]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class st(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "st"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [mstm, "[Mgx]4"],
                [HP_2011_ds62.fst(), "[Fex]4"],
            ],
            energy_interaction=[[16000.0]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class chl(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "chl"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.clin(), "[Mgm][Mgm]4[Alm][Sit1/2Alt1/2]2"],
                [HP_2011_ds62.afchl(), "[Mgm][Mgm]4[Mgm][Sit]2"],
                [HP_2011_ds62.ames(), "[Alm][Mgm]4[Alm][Alt]2"],
                [HP_2011_ds62.daph(), "[Fem][Fem]4[Alm][Sit1/2Alt1/2]2"],
                [ochl1, "[Mgm][Fem]4[Fem][Sit]2"],
                [ochl4, "[Fem][Mgm]4[Mgm][Sit]2"],
            ],
            energy_interaction=[
                [17000.0, 17000.0, 20000.0, 30000.0, 21000.0],
                [16000.0, 37000.0, 20000.0, 4000.0],
                [30000.0, 29000.0, 13000.0],
                [18000.0, 33000.0],
                [24000.0],
            ],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class ctd(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "ctd"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.mctd(), "[Mgmb]"],
                [HP_2011_ds62.fctd(), "[Femb]"],
            ],
            energy_interaction=[[4000.0]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)


class sp1(Solution):
    def __init__(self, molar_fractions=None):
        self.name = "sp1"
        self.solution_model = SymmetricRegularSolution(
            endmembers=[
                [HP_2011_ds62.herc(), "[Fea]"],
                [HP_2011_ds62.sp(), "[Mga]"],
            ],
            energy_interaction=[[0.0]],
        )
        Solution.__init__(self, molar_fractions=molar_fractions)



import numpy as np

try:  # numpy.block was new in numpy version 1.13.0.
    block = np.block(
        [
            [np.ones((3, 3)), 2.0 * np.ones((3, 3))],
            [2.0 * np.ones((3, 3)), 4.0 * np.ones((3, 3))],
        ]
    )
except:
    block = np.array(
        np.bmat(
            [[[[1.0] * 3] * 3, [[2.0] * 3] * 3], [[[2.0] * 3] * 3, [[4.0] * 3] * 3]]
        )
    )
voigt_compliance_factors = block


def voigt_index_to_ij(m):
    """
    Returns the ij (or kl) indices of the
    stiffness tensor which correspond to those
    of the Voigt notation m (or n).
    """
    if m == 3:
        return 1, 2
    elif m == 4:
        return 0, 2
    elif m == 5:
        return 0, 1
    else:
        return m, m


def voigt_notation_to_stiffness_tensor(voigt_notation):
    """
    Converts a stiffness tensor in Voigt notation (6x6 matrix)
    to the full fourth rank tensor (3x3x3x3 matrix).
    """
    stiffness_tensor = np.zeros([3, 3, 3, 3])
    for m in range(6):
        i, j = voigt_index_to_ij(m)
        for n in range(6):
            k, l = voigt_index_to_ij(n)
            stiffness_tensor[i][j][k][l] = voigt_notation[m][n]
            stiffness_tensor[j][i][k][l] = voigt_notation[m][n]
            stiffness_tensor[i][j][l][k] = voigt_notation[m][n]
            stiffness_tensor[j][i][l][k] = voigt_notation[m][n]
    return stiffness_tensor


def voigt_notation_to_compliance_tensor(voigt_notation):
    return voigt_notation_to_stiffness_tensor(
        np.divide(voigt_notation, voigt_compliance_factors)
    )


def contract_stresses(stresses):
    """
    Takes a stress tensor in standard (3x3) form
    and returns the Voigt form (6). No factors
    are required to maintain the relationship
    with the corresponding stiffness and strain tensors
    (see contract_strains, which does require
    multiplicative factors).
    """
    return stresses[[0, 1, 2, 1, 0, 0], [0, 1, 2, 2, 2, 1]]


def expand_stresses(stresses):
    """
    Takes a stress tensor in Voigt form (6)
    and returns the standard form (3x3). No factors
    are required to maintain the relationship
    with the corresponding stiffness and strain tensors
    (see contract_strains, which does require
    multiplicative factors).
    """
    return np.array(
        [
            [stresses[0], stresses[5], stresses[4]],
            [stresses[5], stresses[1], stresses[3]],
            [stresses[4], stresses[3], stresses[2]],
        ]
    )


def contract_strains(strains):
    """
    Takes a stress tensor in standard (3x3) form
    and returns the Voigt form (6). Note the factors
    which are required to maintain the relationship
    with the corresponding stiffness and strain tensors.
    """
    # next line creates a copy, not just a view.
    eps = strains[[0, 1, 2, 1, 0, 0], [0, 1, 2, 2, 2, 1]]
    # only overwrites values in eps, not the input array
    eps[3:] *= 2.0
    return eps


def contract_compliances(compliances):
    """
    Takes a compliance tensor in standard (3x3x3x3) form
    and returns the Voigt form (6x6). Note the compliance factors
    which are required to maintain the inverse relationship with the
    corresponding stiffness tensor.
    """
    voigt_notation = np.zeros((6, 6))
    for p in range(6):
        i, j = voigt_index_to_ij(p)
        for q in range(6):
            m, n = voigt_index_to_ij(q)
            voigt_notation[p, q] = compliances[i, j, m, n]
    return np.multiply(voigt_notation, voigt_compliance_factors)


def contract_stiffnesses(stiffnesses):
    """
    Takes a stiffness tensor in standard (3x3x3x3) form
    and returns the Voigt form (6x6).
    """
    voigt_notation = np.zeros((6, 6))
    for p in range(6):
        i, j = voigt_index_to_ij(p)
        for q in range(6):
            m, n = voigt_index_to_ij(q)
            voigt_notation[p, q] = stiffnesses[i, j, m, n]
    return voigt_notation


def voigt_array_from_cijs(cijs, index_lists):
    """
    Takes a list of cijs and a list of list of tuples corresponding to
    the positions of each cij in the Voigt form matrix.
    Note that the indices run from 0--5, not 1--6.
    """
    C = np.zeros([6, 6])
    for i, index_list in enumerate(index_lists):
        for indices in index_list:
            C[indices] = cijs[i]
            C[indices[::-1]] = cijs[i]
    return C



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np
from scipy.ndimage.filters import gaussian_filter
from scipy.interpolate import RegularGridInterpolator
from sympy import Matrix, Rational
import scipy.integrate as integrate
from collections import Counter
import itertools

from .reductions import row_reduce


def round_to_n(x, xerr, n):
    return round(x, -int(np.floor(np.log10(np.abs(xerr)))) + (n - 1))


def unit_normalize(a, order=2, axis=-1):
    """
    Calculates the L2 normalized array of numpy array a
    of a given order and along a given axis.
    """
    l2 = np.atleast_1d(np.apply_along_axis(np.linalg.norm, axis, a, order))

    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis)[0][0]


def float_eq(a, b):
    """
    Test if two floats are almost equal to each other
    """
    return abs(a - b) < 1e-10 * max(1e-5, abs(a), abs(b))


def linear_interpol(x, x1, x2, y1, y2):
    """
    Linearly interpolate to point x, between
    the points (x1,y1), (x2,y2)
    """
    assert x1 <= x
    assert x2 >= x
    assert x1 <= x2

    alpha = (x - x1) / (x2 - x1)
    return (1.0 - alpha) * y1 + alpha * y2


def bracket(fn, x0, dx, args=(), ratio=1.618, maxiter=100):
    """
    Given a function and a starting guess, find two
    inputs for the function that bracket a root.

    :param fn: The function to bracket.
    :type fn: function
    :param x0: The starting guess.
    :type x0: float
    :param dx: Small step for starting the search.
    :type dx: float
    :param args: Additional arguments to give to fn.
    :type args: tuple
    :param ratio: The step size increases by this ratio
        every step in the search. Defaults to
        the golden ratio.
    :type ratio: float
    :param maxiter: The maximum number of steps before giving up.
    :type maxiter: int

    :returns: xa, xb, fa, fb. xa and xb are the inputs which
        bracket a root of fn. fa and fb are the values of the
        function at those points.
        If the bracket function takes more than maxiter steps,
        it raises a ValueError.
    :rtype: tuple of floats

    """
    niter = 0
    dx = np.abs(dx)
    assert ratio > 1.0

    # Get the starting positions
    f0 = fn(x0, *args)
    x_left = x0 - dx
    x_right = x0 + dx
    f_left = fn(x_left, *args)
    f_right = fn(x_right, *args)

    # Overshot zero, try making dx smaller
    if (f0 - f_left) * (f_right - f0) < 0.0:
        while (
            (f0 - f_left) * (f_right - f0) < 0.0
            and dx > np.finfo("float").eps
            and niter < maxiter
        ):
            dx /= ratio
            x_left = x0 - dx
            x_right = x0 + dx
            f_left = fn(x_left, *args)
            f_right = fn(x_right, *args)
            niter += 1
        if (
            niter == maxiter
        ):  # Couldn't find something with same slope in both directions
            raise ValueError("Cannot find zero.")

    niter = 0
    slope = f_right - f0
    if slope > 0.0 and f0 > 0.0:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left
    elif slope > 0.0 and f0 < 0.0:  # Walk right
        x1 = x_right
        f1 = f_right
    elif slope < 0.0 and f0 > 0:  # Walk right
        x1 = x_right
        f1 = f_right
    else:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left

    # Do the walking
    while f0 * f1 > 0.0 and niter < maxiter:
        dx *= ratio
        xnew = x1 + dx
        fnew = fn(xnew, *args)
        x0 = x1
        f0 = f1
        x1 = xnew
        f1 = fnew
        niter += 1

    if f0 * f1 > 0.0:
        raise ValueError("Cannot find zero.")
    else:
        return x0, x1, f0, f1


def _pad_ndarray_inverse_mirror(array, padding):
    """
    Pads an ndarray according to an inverse mirror
    scheme. For example, for a 1D array
    [2, 4, 6, 7, 8] padded by 3 cells, we have:

     padding  |  original array |  padding

    -3 -2  0  |  2  4  6  7  8  |  9 10 12

    :param array: The array to be padded
    :type array: numpy.ndarray

    :param padding: The number of elements with which to pad the
        array in each dimension.
    :type padding: tuple

    :returns: The padded array.
    :type: numpy.ndarray
    """
    padded_shape = [n + 2 * padding[i] for i, n in enumerate(array.shape)]
    padded_array = np.zeros(padded_shape)

    slices = tuple(
        [slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)]
    )
    padded_array[slices] = array

    padded_array_indices = list(
        itertools.product(
            *[range(n + 2 * padding[i]) for i, n in enumerate(array.shape)]
        )
    )
    inserted_indices = list(
        itertools.product(
            *[range(padding[i], padding[i] + l) for i, l in enumerate(array.shape)]
        )
    )
    padded_array_indices.extend(inserted_indices)

    counter = Counter(padded_array_indices)
    keys = list(counter.keys())
    padded_indices = [keys[i] for i, value in enumerate(counter.values()) if value == 1]
    edge_indices = tuple(
        [
            tuple(
                [
                    np.min(
                        [
                            np.max([axis_idx, padding[dimension]]),
                            padded_array.shape[dimension] - padding[dimension] - 1,
                        ]
                    )
                    for dimension, axis_idx in enumerate(idx)
                ]
            )
            for idx in padded_indices
        ]
    )
    mirror_indices = tuple(
        [
            tuple(
                [
                    2 * edge_indices[i][j] - padded_indices[i][j]
                    for j in range(len(array.shape))
                ]
            )
            for i in range(len(padded_indices))
        ]
    )

    for i, idx in enumerate(padded_indices):
        padded_array[idx] = (
            2.0 * padded_array[edge_indices[i]] - padded_array[mirror_indices[i]]
        )

    return padded_array


def smooth_array(
    array, grid_spacing, gaussian_rms_widths, truncate=4.0, mode="inverse_mirror"
):
    """
    Creates a smoothed array by convolving it with a gaussian filter.
    Grid resolutions and gaussian RMS widths are required for each of
    the axes of the numpy array. The smoothing is truncated at a
    user-defined number of standard deviations. The edges of the array
    can be padded in a number of different ways given by the
    'mode' parameter.

    :param array: The array to smooth.
    :type array: numpy.ndarray
    :param grid_spacing: The spacing of points along each axis.
    :type grid_spacing: numpy.array of floats
    :param gaussian_rms_widths: The Gaussian RMS widths/standard deviations for the
        Gaussian convolution.
    :type gaussian_rms_widths: numpy.array of floats
    :param truncate: The number of standard deviations at which to truncate
        the smoothing.
    :type truncate: float
    :param mode: {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        :func:`burnman.tools.math._pad_ndarray_inverse_mirror`.
    :type mode: str

    :returns: The smoothed array
    :rtype: numpy.ndarray
    """

    # gaussian_filter works with standard deviations normalised to
    # the grid spacing.
    sigma = tuple(np.array(gaussian_rms_widths) / np.array(grid_spacing))

    if mode == "inverse_mirror":
        padding = tuple([int(np.ceil(truncate * s)) for s in sigma])
        padded_array = _pad_ndarray_inverse_mirror(array, padding)
        smoothed_padded_array = gaussian_filter(padded_array, sigma=sigma)
        slices = tuple(
            [slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)]
        )
        smoothed_array = smoothed_padded_array[slices]
    else:
        smoothed_array = gaussian_filter(array, sigma=sigma, mode=mode)

    return smoothed_array


def interp_smoothed_array_and_derivatives(
    array,
    x_values,
    y_values,
    x_stdev=0.0,
    y_stdev=0.0,
    truncate=4.0,
    mode="inverse_mirror",
    indexing="xy",
):
    """
    Creates a smoothed array on a regular 2D grid. Smoothing
    is achieved using :func:`burnman.tools.math.smooth_array`.
    Outputs scipy.interpolate.RegularGridInterpolator() interpolators
    which can be used to query the array, or its derivatives in the
    x- and y- directions.

    :param array: The array to smooth. Each element array[i][j]
        corresponds to the position x_values[i], y_values[j]
    :type array: numpy.array (2D)
    :param x_values: The gridded x values over which to create the smoothed grid
    :type x_values: numpy.array (1D)
    :param y_values: The gridded y_values over which to create the smoothed grid
    :type y_values: numpy.array (1D)
    :param x_stdev: The standard deviation for the Gaussian filter along the x axis
    :type x_stdev: float
    :param y_stdev: The standard deviation for the Gaussian filter along the y axis
    :type y_stdev: float
    :param truncate: The number of standard deviations at which to truncate
        the smoothing.
    :type truncate: float
    :param mode: {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        :func:`burnman.tools.math._pad_ndarray_inverse_mirror`.
    :type mode: str
    :param indexing: {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See numpy.meshgrid for more details.
    :type indexing: str

    :returns: Three RegularGridInterpolator functors
        interpolation functions for the smoothed property and
        the first derivatives with respect to x and y.
    :rtype: tuple
    """

    dx = x_values[1] - x_values[0]
    dy = y_values[1] - y_values[0]

    if indexing == "xy":
        smoothed_array = smooth_array(
            array=array,
            grid_spacing=np.array([dy, dx]),
            gaussian_rms_widths=np.array([y_stdev, x_stdev]),
            truncate=truncate,
            mode=mode,
        ).T

    elif indexing == "ij":
        smoothed_array = smooth_array(
            array=array,
            grid_spacing=np.array([dx, dy]),
            gaussian_rms_widths=np.array([x_stdev, y_stdev]),
            truncate=truncate,
            mode=mode,
        )

    else:
        raise Exception("Indexing scheme not recognised. Should be ij or xy.")

    dSAdxdx, dSAdydy = np.gradient(smoothed_array)

    interps = (
        RegularGridInterpolator((x_values, y_values), smoothed_array, method="linear"),
        RegularGridInterpolator((x_values, y_values), dSAdxdx / dx, method="linear"),
        RegularGridInterpolator((x_values, y_values), dSAdydy / dy, method="linear"),
    )

    return interps


def compare_l2(depth, calc, obs):
    """
    Computes the L2 norm for N profiles at a time (assumed to be linear between points).

    :type depths: array of float
    :param depths: depths. :math:`[m]`
    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: array of L2 norms of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(l2(depth, calc[i], obs[i]))

    return err


def compare_chifactor(calc, obs):
    """
    Computes the chi factor for N profiles at a time. Assumes a 1% a priori uncertainty on the seismic model.


    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: error array of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(chi_factor(calc[i], obs[i]))

    return err


def l2(x, funca, funcb):
    """
    Computes the L2 norm for one profile(assumed to be linear between points).

    :type x: array of float
    :param x: depths :math:`[m]`.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: L2 norm
    :rtype: array of floats
    """
    diff = np.array(funca - funcb)
    diff = diff * diff

    return integrate.trapezoid(diff, x)


def nrmse(x, funca, funcb):
    """
    Normalized root mean square error for one profile
    :type x: array of float
    :param x: depths in m.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: RMS error
    :rtype: array of floats
    """
    diff = np.array(funca - funcb)
    diff = diff * diff
    rmse = np.sqrt(np.sum(diff) / x)
    nrmse = rmse / (np.max(funca) - np.min(funca))

    return nrmse


def chi_factor(calc, obs):
    """
    :math:`\\chi` factor for one profile assuming 1% uncertainty on the reference model (obs)
    :type calc: list of arrays of float
    :param calc: array calculated values
    :type obs: list of arrays of float
    :param obs: array of reference values to compare to

    :returns: :math:`\\chi` factor
    :rtype: array of floats
    """

    err = np.empty_like(calc)
    for i in range(len(calc)):
        err[i] = pow((calc[i] - obs[i]) / (0.01 * np.mean(obs)), 2.0)

    err_tot = np.sum(err) / len(err)

    return err_tot


def independent_row_indices(array):
    """
    Returns the indices corresponding to an independent set of rows
    for a given array. The independent rows are determined from the pivots
    used during row reduction/Gaussian elimination.

    :param array: The input array.
    :type array: 2D numpy.array of floats

    :returns: The indices corresponding to a set of independent rows
        of the input array.
    :rtype: 1D numpy array of integers
    """
    m = Matrix(
        array.shape[0],
        array.shape[1],
        lambda i, j: Rational(array[i, j]).limit_denominator(1000),
    )
    _, pivots, swaps = row_reduce(
        m,
        iszerofunc=lambda x: x.is_zero,
        simpfunc=lambda x: Rational(x).limit_denominator(1000),
    )
    indices = np.array(range(len(array)))
    for swap in np.array(swaps):
        indices[swap] = indices[swap[::-1]]
    return sorted(indices[: len(pivots)])


def array_to_rational_matrix(array):
    """
    Converts a numpy array into a sympy matrix
    filled with rationals
    """
    return Matrix([[Rational(v).limit_denominator(1000) for v in row] for row in array])


def complete_basis(basis, flip_columns=True):
    """
    Creates a full basis by filling remaining rows with
    selected rows of the identity matrix that have indices
    not in the column pivot list of the basis RREF

    :param basis: A partial basis
    :type basis: numpy array
    :param flip_columns: whether to calculate the RREF
        from a flipped basis. This can be useful, for example,
        when dependent columns are known to be further to the
        right in the basis. defaults to True
    :type flip_columns: bool, optional
    :return: basis
    :rtype: numpy array
    """

    n, m = basis.shape
    if n < m:
        if flip_columns:
            pivots = list(m - 1 - np.array(Matrix(np.flip(basis, axis=1)).rref()[1]))
        else:
            pivots = list(Matrix(basis)).rref()[1]

        return np.concatenate(
            (basis, np.identity(m)[[i for i in range(m) if i not in pivots], :]), axis=0
        )
    else:
        return basis


def generate_complete_basis(incomplete_basis, array):
    """
    Given a 2D array with independent rows and a second 2D array that spans a
    larger space, creates a complete basis for the combined array using all
    the rows of the first array, followed by any required rows of the
    second array. So, for example, if the first array is:
    [[1, 0, 0], [1, 1, 0]] and the second array is:
    [[1, 0, 0], [0, 1, 0], [0, 0, 1]], the complete basis will be:
    [[1, 0, 0], [1, 1, 0], [0, 0, 1]].

    :param incomplete_basis: An array containing the basis to be completed.
    :type incomplete_basis: 2D numpy.array

    :param array: An array spanning the full space for which a basis is required.
    :type array: 2D numpy.array

    :returns: An array containing the basis vectors spanning both of the
        input arrays.
    :rtype: 2D numpy array
    """

    incomplete_rank = array_to_rational_matrix(incomplete_basis).rank()
    if incomplete_rank < len(incomplete_basis):
        raise Exception(
            "The incomplete basis is rank-deficient. " "Remove one or more endmembers."
        )

    a = np.concatenate((incomplete_basis, array))
    complete_basis = np.array(a[independent_row_indices(a)], dtype=float)

    # Store the rank of the matrix for later comparison
    len_basis = array_to_rational_matrix(complete_basis).rank()

    # This next step ensures that all of the original
    # rows are contained in the new basis in their original order
    c = np.linalg.lstsq(
        np.array(complete_basis).astype(float).T,
        np.array(incomplete_basis).astype(float).T,
        rcond=None,
    )[0].T

    for row in np.eye(len(c[0])):
        old_rank = array_to_rational_matrix(c).rank()
        c2 = np.concatenate((c, [row]))
        new_rank = array_to_rational_matrix(c2).rank()

        if new_rank > old_rank:
            c = c2

    complete_basis = c.dot(complete_basis)

    # Check that the matrix rank has not changed
    if len_basis != array_to_rational_matrix(complete_basis).rank():
        raise Exception(
            "Basis length changed during conversion. " "Report this bug to developers."
        )

    return complete_basis.round(decimals=12) + 0.0


def is_positive_definite(matrix):
    """
    Checks if a matrix is positive definite

    :param matrix: Input matrix
    :type matrix: 2D numpy array

    :return: Whether or not the matrix is positive definite
    :rtype: bool
    """
    return np.all(np.linalg.eigvals(matrix) > 0)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
Optimization functions
(fitting and solving)
"""

from . import composition_fitting
from . import eos_fitting
from . import linear_fitting
from . import nonlinear_fitting
from . import nonlinear_solvers



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.eos.vinet import Vinet
from burnman.eos import debye
from burnman.classes.calibrant import Calibrant
import numpy as np

"""
Tange_2009
^^^^^^^^^^
"""


class MgO_BM3(Calibrant):
    """
    The MgO pressure standard reported by
    Tange et al. (2009; https://doi.org/10.1029/2008JB005813).
    BM3 version.
    """

    def __init__(self):
        def _pressure_Tange_MgO(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            gra = params["grueneisen_0"] * (
                1.0 + params["a"] * ((volume / params["V_0"]) ** params["b"] - 1.0)
            )
            debye_Ta = (
                params["Debye_0"]
                * (volume / params["V_0"])
                ** (-params["grueneisen_0"] * (1.0 - params["a"]))
                * np.exp(-(gra - params["grueneisen_0"]) / params["b"])
            )

            E_th = debye.thermal_energy(temperature, debye_Ta, params["n"])
            E_th_ref = debye.thermal_energy(
                params["T_0"], debye_Ta, params["n"]
            )  # thermal energy at reference temperature

            Pth = gra * (E_th - E_th_ref) / volume

            # Total pressure
            P = P0 + Pth

            return P

        _params_Tange_MgO = {
            "V_0": 1.12463e-05,
            "K_0": 160.64e9,
            "Kprime_0": 4.221,
            "Debye_0": 761.0,
            "grueneisen_0": 1.431,
            "a": 0.29,
            "b": 3.5,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Tange_MgO, "pressure", _params_Tange_MgO)


class MgO_Vinet(Calibrant):
    """
    The MgO pressure standard reported by
    Tange et al. (2009; https://doi.org/10.1029/2008JB005813).
    Vinet version.
    """

    def __init__(self):
        def _pressure_Tange_MgO(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = Vinet()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            gra = params["grueneisen_0"] * (
                1.0 + params["a"] * ((volume / params["V_0"]) ** params["b"] - 1.0)
            )
            debye_Ta = (
                params["Debye_0"]
                * (volume / params["V_0"])
                ** (-params["grueneisen_0"] * (1.0 - params["a"]))
                * np.exp(-(gra - params["grueneisen_0"]) / params["b"])
            )

            E_th = debye.thermal_energy(temperature, debye_Ta, params["n"])
            E_th_ref = debye.thermal_energy(
                params["T_0"], debye_Ta, params["n"]
            )  # thermal energy at reference temperature

            Pth = gra * (E_th - E_th_ref) / volume

            # Total pressure
            P = P0 + Pth

            return P

        _params_Tange_MgO = {
            "V_0": 1.12463e-05,
            "K_0": 160.63e9,
            "Kprime_0": 4.367,
            "Debye_0": 761.0,
            "grueneisen_0": 1.442,
            "a": 0.138,
            "b": 5.4,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Tange_MgO, "pressure", _params_Tange_MgO)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Sundman 1991 models for iron
Written in the Holland and Powell, 2011 form

EoS terms for bcc are from HP_2011_ds62 for iron
EoS terms for fcc are from an unpublished calibration
(R. Myhill, 02/2017)
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..utils.chemistry import formula_mass

"""
ENDMEMBERS
"""


class bcc_iron(Mineral):
    def __init__(self):
        formula = {"Fe": 1.0}
        self.params = {
            "name": "BCC iron",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": 9149.0,
            "S_0": 36.868,
            "V_0": 7.09e-06,
            "Cp": [21.09, 0.0101455, -221508.0, 47.1947],
            "a_0": 3.56e-05,
            "K_0": 1.64e11,
            "Kprime_0": 5.16,
            "Kdprime_0": -3.1e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        self.property_modifiers = [
            [
                "magnetic_chs",
                {
                    "structural_parameter": 0.4,
                    "curie_temperature": [1043.0, 0.0],
                    "magnetic_moment": [2.22, 0.0],
                },
            ]
        ]
        Mineral.__init__(self)


class fcc_iron(Mineral):
    def __init__(self):
        formula = {"Fe": 1.0}
        self.params = {
            "name": "FCC iron",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": 7973.0,
            "S_0": 35.907,
            "V_0": 6.93863394593e-06,
            "Cp": [22.24, 0.0088656, -221517.0, 47.1998],
            "a_0": 5.13e-05,
            "K_0": 1.539e11,
            "Kprime_0": 5.2,
            "Kdprime_0": -3.37e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }
        self.property_modifiers = [
            [
                "magnetic_chs",
                {
                    "structural_parameter": 0.28,
                    "curie_temperature": [201.0, 0.0],
                    "magnetic_moment": [2.1, 0.0],
                },
            ]
        ]
        Mineral.__init__(self)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Murakami_etal_2012
^^^^^^^^^^^^^^^^^^

Minerals from Murakami et al. (2012) supplementary table 5 and references therein, V_0 from
Stixrude & Lithgow-Bertolloni 2005. Some information from personal communication with Murakami.


"""
from __future__ import absolute_import
from ..classes import mineral_helpers as helpers
from ..classes.mineral import Mineral


class mg_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 24.45e-6,  # S & L-B 2005
            "K_0": 281e9,
            "Kprime_0": 4.1,
            "G_0": 173e9,
            "Gprime_0": 1.56,
            "molar_mass": 0.100,
            "n": 5,
            "Debye_0": 1070.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
            "eta_s_0": 2.4,
        }

        Mineral.__init__(self)


class mg_perovskite_3rdorder(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 24.45e-6,  # S & L-B 2005
            "K_0": 281e9,
            "Kprime_0": 4.1,
            "G_0": 171.42e9,
            "Gprime_0": 1.83,
            "molar_mass": 0.100,
            "n": 5,
            "Debye_0": 1070.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
            "eta_s_0": 2.4,
        }

        Mineral.__init__(self)


class fe_perovskite(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 24.607e-6,
            "K_0": 251.9e9,
            "Kprime_0": 4.01,
            "G_0": 164.7e9,
            "Gprime_0": 1.58,
            "molar_mass": 0.102,
            "n": 5,
            "Debye_0": 1054.0,
            "grueneisen_0": 1.48,
            "q_0": 1.4,
            "eta_s_0": 2.4,
        }

        Mineral.__init__(self)


class mg_periclase(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 11.24e-6,  # S & L-B 2005
            "K_0": 161e9,
            "Kprime_0": 3.9,
            "G_0": 131e9,
            "Gprime_0": 1.92,
            "molar_mass": 0.0403,
            "n": 2,
            "Debye_0": 773.0,  # S& L-B 2005
            "grueneisen_0": 1.5,
            "q_0": 1.5,  # S&L-B 2005
            "eta_s_0": 3.0,
        }

        Mineral.__init__(self)


class fe_periclase(helpers.HelperSpinTransition):
    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 63.0e9, fe_periclase_LS(), fe_periclase_HS()
        )


class fe_periclase_3rd(helpers.HelperSpinTransition):
    def __init__(self):
        helpers.HelperSpinTransition.__init__(
            self, 63.0e9, fe_periclase_LS(), fe_periclase_HS()
        )


class fe_periclase_HS(
    Mineral
):  # From Murakami's emails, see Cayman for details, represents Mg# = .79
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 11.412e-6,
            "K_0": 159.1e9,
            "Kprime_0": 4.11,
            "G_0": 105.43e9,
            "Gprime_0": 1.773,
            "molar_mass": 0.047,
            "n": 2,
            "Debye_0": 706.0,
            "grueneisen_0": 1.45,
            "q_0": 1.5,
            "eta_s_0": 2.54,
        }

        Mineral.__init__(self)


class fe_periclase_LS(
    Mineral
):  # From Murakami's emails, see Cayman for details, represents Mg# = .79
    def __init__(self):
        self.params = {
            "equation_of_state": "slb2",
            "V_0": 11.171e-6,
            "K_0": 170.0e9,
            "Kprime_0": 4.00,
            "G_0": 116.34e9,
            "Gprime_0": 1.668,
            "molar_mass": 0.047,
            "n": 2,
            "Debye_0": 706.0,
            "grueneisen_0": 1.45,
            "q_0": 1.5,
            "eta_s_0": 2.54,
        }

        Mineral.__init__(self)


class fe_periclase_HS_3rd(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 11.412e-6,
            "K_0": 159.1e9,
            "Kprime_0": 4.11,
            "G_0": 129.35e9,
            "Gprime_0": 1.993,
            "molar_mass": 0.0469,
            "n": 2,
            "Debye_0": 706.0,
            "grueneisen_0": 1.45,
            "q_0": 1.5,
            "eta_s_0": 2.54,
        }

        Mineral.__init__(self)


class fe_periclase_LS_3rd(Mineral):
    def __init__(self):
        self.params = {
            "equation_of_state": "slb3",
            "V_0": 11.171e-6,
            "K_0": 170.0e9,
            "Kprime_0": 4.0,
            "G_0": 151.67e9,
            "Gprime_0": 1.754,
            "molar_mass": 0.0469,
            "n": 2,
            "Debye_0": 706.0,
            "grueneisen_0": 1.45,
            "q_0": 1.5,
            "eta_s_0": 2.54,
        }

        Mineral.__init__(self)


mg_bridgmanite = mg_perovskite
fe_bridgmanite = fe_perovskite
mg_bridgmanite_3rdorder = mg_perovskite_3rdorder



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.classes.calibrant import Calibrant
import numpy as np


"""
LeGodec_2000
^^^^^^^^^^^^
"""


class hBN(Calibrant):
    """
    The hexagonal boron nitride pressure standard reported by
    LeGodec et al. (2000; http://dx.doi.org/10.1080/08957950008200304).
    """

    def __init__(self):
        def _pressure_LeGodec_hBN(volume, temperature, params):

            # Modified BM3+Thermal (see Zhao et al. 1997)
            a0 = params["a0"]
            a1 = params["a1"]
            K0 = params["K_0"]
            Kprime0 = params["Kprime_0"]
            dK_dT = params["dK_dT"]
            dKprime_dT = params["dKprime_dT"]

            KT = (K0) + dK_dT * (temperature - params["T_0"])
            KprimeT = Kprime0 + dKprime_dT * (temperature - params["T_0"])
            V0T = params["V_0"] * np.exp(
                ((a0 * temperature) + (a1 * ((temperature**2.0) / 2.0)))
                - ((a0 * params["T_0"]) + (a1 * ((params["T_0"] ** 2.0) / 2.0)))
            )
            V_V0T = volume / V0T
            f = 0.5 * ((V_V0T ** (-2.0 / 3.0)) - 1.0)

            P = (3.0 * KT * f * (1.0 + 2.0 * f) ** (5.0 / 2.0)) * (
                1.0 - 3.0 / 2.0 * (4.0 - KprimeT) * f
            )

            return P

        _params_LeGodec_hBN = {
            "V_0": 1.0891e-05,
            "K_0": 27.6e9,
            "Kprime_0": 10.5,
            "a0": 3.53e-05,
            "a1": 2.1e-09,
            "dK_dT": -0.0081e9,
            "dKprime_dT": 0.0016,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 2.0,
        }

        Calibrant.__init__(self, _pressure_LeGodec_hBN, "pressure", _params_LeGodec_hBN)



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from . import equation_of_state as eos
from ..utils.math import bracket
import warnings


def bulk_modulus(volume, params):
    """
    compute the bulk modulus as per the third order
    birch-murnaghan equation of state.  Returns bulk
    modulus in the same units as the reference bulk
    modulus.  Pressure must be in :math:`[Pa]`.
    """

    x = params["V_0"] / volume
    f = 0.5 * (pow(x, 2.0 / 3.0) - 1.0)

    K = pow(1.0 + 2.0 * f, 5.0 / 2.0) * (
        params["K_0"]
        + (3.0 * params["K_0"] * params["Kprime_0"] - 5 * params["K_0"]) * f
        + 27.0
        / 2.0
        * (params["K_0"] * params["Kprime_0"] - 4.0 * params["K_0"])
        * f
        * f
    )
    return K


def birch_murnaghan(x, params):
    """
    equation for the third order birch-murnaghan equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    """

    return (
        3.0
        * params["K_0"]
        / 2.0
        * (pow(x, 7.0 / 3.0) - pow(x, 5.0 / 3.0))
        * (1.0 - 0.75 * (4.0 - params["Kprime_0"]) * (pow(x, 2.0 / 3.0) - 1.0))
        + params["P_0"]
    )


def volume(pressure, params):
    """
    Get the birch-murnaghan volume at a reference temperature for a given
    pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
    """

    func = lambda x: birch_murnaghan(params["V_0"] / x, params) - pressure
    try:
        sol = bracket(func, params["V_0"], 1.0e-2 * params["V_0"])
    except:
        raise ValueError(
            "Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?"
        )
    return opt.brentq(func, sol[0], sol[1])


def shear_modulus_second_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in :math:`[Pa]` (the same units as in
    params['G_0']).  This uses a second order finite strain expansion
    """

    x = params["V_0"] / volume
    G = (
        params["G_0"]
        * pow(x, 5.0 / 3.0)
        * (
            1.0
            - 0.5
            * (pow(x, 2.0 / 3.0) - 1.0)
            * (5.0 - 3.0 * params["Gprime_0"] * params["K_0"] / params["G_0"])
        )
    )
    return G


def shear_modulus_third_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in :math:`[Pa]` (the same units as in
    params['G_0']).  This uses a third order finite strain expansion
    """

    x = params["V_0"] / volume
    f = 0.5 * (pow(x, 2.0 / 3.0) - 1.0)
    G = pow((1.0 + 2.0 * f), 5.0 / 2.0) * (
        params["G_0"]
        + (3.0 * params["K_0"] * params["Gprime_0"] - 5.0 * params["G_0"]) * f
        + (
            6.0 * params["K_0"] * params["Gprime_0"]
            - 24.0 * params["K_0"]
            - 14.0 * params["G_0"]
            + 9.0 / 2.0 * params["K_0"] * params["Kprime_0"]
        )
        * f
        * f
    )
    return G


class BirchMurnaghanBase(eos.EquationOfState):
    """
    Base class for the isothermal Birch Murnaghan equation of state.  This is third order in strain, and
    has no temperature dependence.  However, the shear modulus is sometimes fit to a second order
    function, so if this is the case, you should use that.  For more see :class:`burnman.birch_murnaghan.BM2` and :class:`burnman.birch_murnaghan.BM3`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return birch_murnaghan(params["V_0"] / volume, params)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        if self.order == 2:
            return shear_modulus_second_order(volume, params)
        elif self.order == 3:
            return shear_modulus_third_order(volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = np.power(volume / params["V_0"], -1.0 / 3.0)
        x2 = x * x
        x4 = x2 * x2
        x6 = x4 * x2
        x8 = x4 * x4

        xi1 = 3.0 * (4.0 - params["Kprime_0"]) / 4.0

        intPdV = (
            -9.0
            / 2.0
            * params["V_0"]
            * params["K_0"]
            * (
                (xi1 + 1.0) * (x4 / 4.0 - x2 / 2.0 + 1.0 / 4.0)
                - xi1 * (x6 / 6.0 - x4 / 4.0 + 1.0 / 12.0)
            )
        )

        return -intPdV + params["E_0"]

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + volume * pressure
        )

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 0.0

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if "G_0" not in params:
            params["G_0"] = float("nan")
        if "Gprime_0" not in params:
            params["Gprime_0"] = float("nan")

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0", "G_0", "Gprime_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 20.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["G_0"] < 0.0 or params["G_0"] > 1.0e13:
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0:
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)


class BM3(BirchMurnaghanBase):
    """
    Third order Birch Murnaghan isothermal equation of state.
    This uses the third order expansion for shear modulus.
    """

    def __init__(self):
        self.order = 3


class BM2(BirchMurnaghanBase):
    """
    Third order Birch Murnaghan isothermal equation of state.
    This uses the second order expansion for shear modulus.
    """

    def __init__(self):
        self.order = 2



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import warnings
import numpy as np

from . import equation_of_state as eos


def tait_constants(params):
    """
    returns parameters for the modified Tait equation of state
    derived from K_T and its two first pressure derivatives
    EQ 4 from Holland and Powell, 2011
    """
    a = (1.0 + params["Kprime_0"]) / (
        1.0 + params["Kprime_0"] + params["K_0"] * params["Kdprime_0"]
    )
    b = params["Kprime_0"] / params["K_0"] - params["Kdprime_0"] / (
        1.0 + params["Kprime_0"]
    )
    c = (1.0 + params["Kprime_0"] + params["K_0"] * params["Kdprime_0"]) / (
        params["Kprime_0"] * params["Kprime_0"]
        + params["Kprime_0"]
        - params["K_0"] * params["Kdprime_0"]
    )
    return a, b, c


def pressure_modified_tait(Vrel, params):
    """
    Pressure according to the modified Tait equation of state.
    EQ 2 from Holland and Powell, 2011

    :param Vrel: V/V_0
    :type Vrel: float or numpy array
    :param params: Parameter dictionary
    :type params: dictionary
    :return: pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    :rtype: float or numpy array
    """
    a, b, c = tait_constants(params)
    return (np.power((Vrel + a - 1.0) / a, -1.0 / c) - 1.0) / b + params["P_0"]


def volume(pressure, params):
    """
    Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
    EQ 12
    """
    a, b, c = tait_constants(params)
    Vrel = 1.0 - a * (1.0 - np.power((1.0 + b * (pressure - params["P_0"])), -1.0 * c))
    return Vrel * params["V_0"]


def bulk_modulus(pressure, params):
    """
    Returns isothermal bulk modulus :math:`K_T` of the mineral. :math:`[Pa]`.
    EQ 13+2
    """
    a, b, c = tait_constants(params)
    return (
        params["K_0"]
        * (1.0 + b * (pressure - params["P_0"]))
        * (a + (1.0 - a) * np.power((1.0 + b * (pressure - params["P_0"])), c))
    )


def intVdP(pressure, params):
    """
    Returns the integral of VdP for the mineral. :math:`[J]`.
    EQ 13
    """
    a, b, c = tait_constants(params)
    psubpth = pressure - params["P_0"]

    if pressure != params["P_0"]:
        intVdP = (
            (pressure - params["P_0"])
            * params["V_0"]
            * (
                1.0
                - a
                + (
                    a
                    * (1.0 - np.power((1.0 + b * (psubpth)), 1.0 - c))
                    / (b * (c - 1.0) * (pressure - params["P_0"]))
                )
            )
        )
    else:
        intVdP = 0.0
    return intVdP


class MT(eos.EquationOfState):
    """
    Base class for the generic modified Tait equation of state.
    References for this can be found in :cite:`HC1974`
    and :cite:`HP2011` (followed here).

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('mt')
    (or by initialising the class with the param
    equation_of_state = 'mt').
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        return pressure_modified_tait(volume / params["V_0"], params)

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(pressure, params)

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[Pa]`
        """
        return 1.0e99

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented in the Modified Tait EoS. :math:`[Pa]`
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.0

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """

        return (
            self.gibbs_free_energy(pressure, temperature, volume, params)
            - volume * pressure
        )

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV
        a, b, c = tait_constants(params)

        intVdP = params["V_0"] * (
            a
            / (b * (1.0 - c))
            * (np.power(b * (pressure - params["P_0"]) + 1.0, 1.0 - c) - 1.0)
            + (1.0 - a) * (pressure - params["P_0"])
        )

        return intVdP + params["E_0"] + params["V_0"] * params["P_0"]

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.0e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 1.0e5

        # G and Gprime are not defined in this equation of state,
        # We can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if "G_0" not in params:
            params["G_0"] = float("nan")
        if "Gprime_0" not in params:
            params["Gprime_0"] = float("nan")

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0", "Kdprime_0", "G_0", "Gprime_0"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-2:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 40.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["G_0"] < 0.0 or params["G_0"] > 1.0e13:
            warnings.warn("Unusual value for G_0", stacklevel=2)
        if params["Gprime_0"] < -5.0 or params["Gprime_0"] > 10.0:
            warnings.warn("Unusual value for Gprime_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import importlib
import numpy as np
from scipy.linalg import inv, sqrtm
import warnings

try:
    cp = importlib.import_module("cvxpy")
except ImportError as err:
    print(
        f"Warning: {err}. " "For full functionality of BurnMan, please install cvxpy."
    )


def weighted_constrained_least_squares(
    A, b, Cov_b=None, equality_constraints=None, inequality_constraints=None
):
    """
    Solves a weighted, constrained least squares problem using cvxpy.
    The objective function is to minimize the following:
    sum_squares(Cov_b^(-1/2).A.x - Cov_b^(-1/2).b))
    subject to
    C.x == c
    D.x <= d

    :param A: An array defining matrix A in the objective function above.
    :type A: 2D numpy array

    :param b: An array defining vector b in the objective function above.
    :type b: numpy array

    :param Cov_b: A covariance matrix associated with b
    :type Cov_b: 2D numpy array

    :param equality_constraints: A list containing the matrices C and c
        in the objective function above.
    :type equality_constraints: list containing a 2D array and 1D array

    :param inequality_constraints: A list containing the matrices D and d
        in the objective function above.
    :type inequality_constraints: list containing a 2D array and 1D array

    :returns: Tuple containing the optimized phase amounts (1D numpy.array),
        a covariance matrix corresponding to the optimized phase amounts
        (2D numpy.array), and the weighted residual of the fitting procedure
        (a float).
    :rtype: tuple
    """

    if Cov_b is None:
        Cov_b = np.eye(len(b))

    # Create the standard weighted least squares objective function
    # (https://stats.stackexchange.com/a/333551)
    n_vars = A.shape[1]
    m = inv(sqrtm(Cov_b))
    mA = m @ A
    mb = m @ b
    x = cp.Variable(n_vars)
    objective = cp.Minimize(cp.sum_squares(mA @ x - mb))

    constraints = []
    if equality_constraints is not None:
        n_eq_csts = len(equality_constraints[0])
        constraints = [
            equality_constraints[0][i] @ x == equality_constraints[1][i]
            for i in range(n_eq_csts)
        ]

    if inequality_constraints is not None:
        n_ineq_csts = len(inequality_constraints[0])
        constraints.extend(
            [
                inequality_constraints[0][i] @ x <= inequality_constraints[1][i]
                for i in range(n_ineq_csts)
            ]
        )

    # Set up the problem and solve it
    warns = []
    if len(constraints) > 1:
        prob = cp.Problem(objective, constraints)
    else:
        prob = cp.Problem(objective)

    try:
        with warnings.catch_warnings(record=True) as w:
            res = prob.solve()
            popt = np.array([x.value[i] for i in range(len(A.T))])
            warns.extend(w)
    except Exception:
        try:
            with warnings.catch_warnings(record=True) as w:
                res = prob.solve(solver=cp.ECOS)
                popt = np.array([x.value[i] for i in range(len(A.T))])
                warns.extend(w)
        except Exception as e:
            raise Exception(e)

    # Calculate the covariance matrix
    # (also from https://stats.stackexchange.com/a/333551)
    inv_Cov_b = np.linalg.inv(Cov_b)
    pcov = np.linalg.inv(A.T.dot(inv_Cov_b.dot(A)))

    return (popt, pcov, res)
